/* segment chanfunc in program nc */

/* routines to set up channel parameters */

#include <stdio.h>
#include "nc.h"
#include "y.tab.h"
#include "ncsub.h"
#include "ncomp.h"
#include "control.h"
#include "ncelem.h"

static Symbol *fpnt;

static int nchantype = 0;
static chantype chtypes[NCHANCONST];
chantype *findchantype(int ctype, int stype);
void initchan(void);

#ifdef __cplusplus
extern "C" {
#endif

double exp(double);
double log(double);
double log10(double);

#ifdef __cplusplus
}
#endif

double ncabs(double x);
double callfunc(Symbol *funcp, int npar, double par1, double par2,
					 double par3, double par4);
char *emalloc(unsigned int n);
char *findsym(int);

/*----------------------------------------*/

chantype *makchantype(int ctype, int stype, int nstates, int nparm, 
				double basetempc)

/* Information describing the state variables of a HH-type channel
  (e.g. m or h, etc.) or a Markov sequential-state channel.

  Contains table to hold values of alpha and beta calculated by
  the same function call, and with the same Q10, table to hold
  value of m1, m2 for implicit integration of the variable, pointer
  to default function to calculate rates, name of user-defined
  function to override calculation of rates, and storage space for
  the Q10 value, v, timinc, and tempcel variables used to
  determine at run time if table calculation is still valid.
*/

{
   chantype *ch;
   chanparm *chp;
   int i;

/*
  if (!(ch=(chanstate *)emalloc(sizeof(chantype)))) {
     fprintf (stderr,"makchantype: can't allocate table of size %d\n",nstates);
     return ((chantype*)NULL);
  }
*/

  if  (!(ch=findchantype(ctype,stype))) {
      ch = &chtypes[nchantype++];		/* allocate new chan type */
  }
  if (nchantype >= NCHANCONST) {
	nchantype = NCHANCONST;
	fprintf (stderr,"makchantype: too many chan types\n");
  }
 
  ch->ctype = ctype;
  ch->stype  = stype;

  ch->numstate = nstates;
  ch->numparm = nparm;		 
  ch->hh = 0;		 	/* = 1 -> HH implicit tables used for rates */
  ch->qc = dqc;			/* Q10 factor for conductance */
  ch->qt = basetempc;		/* base temperature for channel rate & cond */
  ch->qcond = 0;		/* Q10 for unitary conductance */
  ch->vrev = NULLVAL;
  ch->perm = (double *)NULL;
  ch->unitary = 6.0e-12;
  ch->trconc = dstr;
  ch->state = (chanstate *)NULL;
  ch->parm = (chanparm *)NULL;
  ch->respamp = (float *)NULL;
  ch->nvval  = 0;
  ch->chanvtab = (float *)NULL;

  if (nstates) {		/* allocate space for state definitions */
    if (nstates > NUMSTATE) {
	fprintf (stderr,"makchantype: %d too many states\n",nstates);
        nstates = NUMSTATE;
    }
    if (!(ch->state=(chanstate *)emalloc(nstates*sizeof(chanstate)))) {
      fprintf(stderr,"makchantype: can't allocate table of size %d\n",nstates);
       return ((chantype*)NULL);
    }
  }
				/* allocate space for param table */
  if (!(ch->parm=(chanparm *)emalloc(nparm*sizeof(chanparm)))) {
    fprintf (stderr,"Initchanparm: can't allocate table of size %d\n",nparm);
    return ((chantype*)NULL);
  }
  for (i=0; i<nparm; i++) {
    chp = &ch->parm[i];
    chp->pn  = i;
    chp->nfval  = 0;
    chp->nival  = 0;
    chp->fval  = (double *)NULL;
    chp->ival  = (double *)NULL;
    chp->chanftab = (float *)NULL;
    chp->chanitab = (float *)NULL;
    chp->chancalc = (double (*)(double,int))NULL;
    chp->funcname = (char *)NULL;
    chp->qrate = 1.0;		/* Q val calculated from qt, dq */
    chp->dq = 1.0;		/* Q10 temp factor for param */
    chp->qt = ch->qt;		/* base temperature for rate calc */
    chp->dqca = dcavoff;	/* Q10 ca voffset factor for param */
    chp->qcao = DCAO + DMGO;	/* base cao, mgo for voffset calc */
    chp->qcavoff = 0;		/* voffset calc from dqca, dcao,dcoo,dmgo */
    chp->voff = 0;		/* voffset from user */
    chp->oldvf = -1e10;
    chp->oldvi = -1e10;
    chp->oldmr = -1e10;
    chp->oldtimincf = 0;
    chp->oldtiminci = 0;
    chp->oldtempcel = 0;
    chp->oldcao = 0;
  }
  if (ctype >= GLU && ctype <= CGMP) {	   /* if receptor binds ligand */
    if (!(ch->respamp=(float *)emalloc(NNTR*sizeof(float)))) {
      fprintf(stderr,"makchantype: can't allocate respamp of size %d\n",NNTR);
      return ((chantype*)NULL);
    }
    for (i=0; i<NNTR; i++) {
      ch->respamp[i] = 0;
    }
  }
  if (!(ch->perm=(double  *)emalloc(NPERM*sizeof(double)))) {
    fprintf(stderr,"makchantype: can't allocate perm of size %d\n",NPERM);
    return ((chantype*)NULL);
  }
  for (i=0; i<NPERM; i++) {
     ch->perm[i] = 0;
  }
  return ch;
}

/*--------------------------------------------*/

double qrate (chanparm *chp)

/* Compute Q10 for channel rate functions */

/* Return rate function multiplier which depends on temperature.
   Base for HH rate functions = 6.3 deg C.

  Assumes that timinc and/or tempcel may change during a run,
    but that the Q10 value and base temperature do not.
    Function of time step.

     chp->dq = Q10 value
     chp->qt = base temperature for calculation
*/

{
  if (tempcel != chp->oldtempcel) {

    chp->qrate = exp(log(chp->dq) * (tempcel - chp->qt) / 10.0);
    chp->oldtempcel = tempcel;
  }
  return (chp->qrate * timinc);
}

/*--------------------------------------------*/

double qcond (chantype *ch)

/* Compute Q10 multiplier for channel unitary conductance */

/* Return conductance multiplier which depends on temperature.
   Base for HH rate functions = 6.3 deg C.

     ch->qc    = Q10 value
     ch->qt    = base temperature for calculation
     ch->qcond = multiplier calculated from qc, qt.
*/

{
  ch->qcond = exp(log(ch->qc) * (tempcel - ch->qt) / 10.0);
  return ch->qcond;
}

/*--------------------------------------------*/

double qcavoff (chanparm *chp)

/* Compute Q10 for Cao effect on channel voltage offset */

/* Return voffset which depends on cao. */

{
  if (dcao != chp->oldcao) {
    chp->qcavoff = chp->dqca * log10( (dcao+dmgo+dcoo) / chp->qcao);
    chp->oldcao = dcao + dmgo + dcoo;
  }
  return chp->qcavoff;
}

/*----------------------------------------*/

chantype *findchantype(int ctype, int stype)

{
   int i,found;
   chantype *ch;

  for (found=i=0; !found && i<nchantype; i++) {
    if ((chtypes[i].ctype == ctype) && (chtypes[i].stype == stype)) {
      ch = &chtypes[i];
      found = 1;
      break;
    }
  }
  if (!found) ch = ((chantype *)NULL);
  return ch;
}

/*----------------------------------------*/

chantype *getchantype(int ctype, int stype)

{
     chantype *ch;
     int found;

  found = 1;
  if (!(ch=findchantype(ctype,stype))) {
     initchan();
     if (!(ch=findchantype(ctype,stype))) {
	found = 0;
	if (stype==0) {		/* default type not found, try type 1 */
	   found = 1;
           if (!(ch=findchantype(ctype,1))) found=0; 
	}
     }
  }
  if (!found) {
   fprintf(stderr,"## getchantype: don't know how to make chan '%s type %d'\n",
		 findsym(ctype),stype);
    ch  = ((chantype *)NULL); 
  }
  return ch;
}

/*----------------------------------------*/

double callchancalc (double v, int func)

/* Call a user-defined function for calculating chan rate functions. */

{
   return (callfunc (fpnt,2,v,(double)func,0,0));
}

/*----------------------------------------*/

void makchanratetab(chanparm *chp)

/* Make tables for rate functions (e.g. alpha and beta) for any 
   standard ("HH") or Markov sequential-state channel type.
   Makes "ntab" tables (e.g. am, bm, etc.).

   Hodgkin-Huxley equations for rate constants are calculated
   from the negative of the difference between the voltage and the
   resting potential.  The tables are calculated here from -250 to
   250 mv (0 means -250 mv, 500 means 250 mv), and are calibrated in
   units of per second.  A final multiplication by "timinc"
   calibrates the tables in terms of the time step.  Therefore, the
   tables need to be recalculated every time the time step changes.

   Note that the rate function called here starts with the first 
   rate func = 1, but that the rate table starts at 0.

   Note that am and bm rate functions are calculated with a
   temperature dependence factor of q.  Normally this represents a
   Q10 of 3, but am and bm are calculated with a Q10 of 2
   (Frankenhaeuser and Moore, 1963, J Physiol 169:431-467).  This
   temperature dependence is implemented in the function qrate();

*/

{
  double v,tval,q;
  float *npt,*chantab;
  double *fval, voff;
  int vn,t,nfval,ratesiz;
  Symbol *f,*lookup(char *str);
  double (*chanf)(double,int);

  static float *am,*bm;

  static float *m1tab;
  static float *m2tab;

/* fprintf (stderr,"maktabk begin\n"); /* */

  nfval = (int)chp->nfval;
  if (nfval < 2) nfval = 2;	/* always make space for at least 2 values */

  fval = chp->fval;
  if (!fval) {		
        if (!(fval=(double *) emalloc(nfval * sizeof(double)))) {
          fprintf (stderr, "Makchanstatetab: can't allocate table for %s\n",
	 		chp->funcname);
        return;
     }
    chp->fval = fval;
    if (chp->nfval < 1) {	/* make dummy values in case no rate function */
      chp->fval[0] = 1.0;
      chp->fval[1] = 1.0;
    }
  }
			/* Don't make voltage table if none specified */

  if ((nfval=(int) chp->nfval) < 1) return; 

  if (!(f=lookup(chp->funcname))) chanf = chp->chancalc;
  else {
     fpnt = f;
     chanf = callchancalc;
  }

  if (!chanf) return;	/* if no rate function defined, return */
 
  chantab = chp->chanftab;
  if (!chantab) {		
        if (!(chantab=(float *)		
                emalloc(nfval * (RATESIZ+1) * sizeof(float)))) {
        fprintf (stderr, "Makchanstatetab: can't allocate table for %s\n",
	 		chp->funcname);
        return;
     }
    chp->chanftab = chantab;
  }
 q = qrate(chp);	/* set Q10 chtyp, func of V and temperature */
 ratesiz = RATESIZ+1;
 npt = chantab;
 for (t=1; t<=nfval; t++) {  /* calc tables for alpha, beta rate funcs */ 
				 /*  and their m1tab, m2tab lookups */
  for (vn=0; vn<=RATESIZ; vn++) {   /* make tables from -250 to 250 mV, */
                                    /*  i.e. vn==0 -> Vm= -250 mV */
                                    /* v = Vm, calib in mV */
    v = vn - (RATESIZ/2) - (chp->voff + qcavoff(chp)) * MVOLT; 
    *(npt + vn) = q * (*chanf)(v,t);

  /*  fprintf (stderr,"func %d V %g rate %g\n", t, v, *(npt+vn)); /* */

  }  /* for (vn;;) */
  npt += ratesiz;

 }  /* for (t;;) */

 chp->oldvf = -1e10;
}

/*----------------------------------------*/

void makchanimpltab(chanparm *chp)

/* Make implicit tables for rate functions alpha and beta for any 
   standard ("HH") channel type.  Makes 2 tables (nival=2), m1tab, m2tab.

   Offset table voltage by default offset value.
*/

{
  float *npt,*chanftab,*chanitab;
  float *am,*bm,*m1tab,*m2tab;
  double *ival,tval;
  int nival,ratesiz,vn;

  nival = (int)chp->nival;
  if (nival < 2) nival = 2;	/* always make space for at least 2 values */


  ival = chp->ival;
  if (!ival) {		
        if (!(ival=(double *) emalloc(nival * sizeof(double)))) {
          fprintf (stderr, "Makchanstatetab: can't allocate table for %s\n",
	 		chp->funcname);
        return;
     }
    chp->ival = ival;
  }
			/* Don't make voltage table if none specified */
  if ((nival=(int) chp->nival) < 2) return; 

  chanitab = chp->chanitab;
  if (!chanitab) {		
        if (!(chanitab=(float *)		
                emalloc(nival * (RATESIZ+1) * sizeof(float)))) {
        fprintf (stderr, "Makchanimpltab: can't allocate table for %s\n",
	 		chp->funcname);
        return;
     }
     chp->chanitab = chanitab;
  }

  ratesiz = RATESIZ+1;
  chanftab = chp->chanftab;
  am    = chanftab;			/* alpha, beta tables, already done */
  bm    = chanftab + ratesiz;

  m1tab = chanitab;			/* implicit tables */
  m2tab = chanitab + ratesiz;

  if (am && m1tab) { 

		/* next, calculate implicit tables */

   for (vn=0; vn<=RATESIZ; vn++) {   /* make tables from -250 to 250 mV, */

    tval  =	  (1.0 - (am[vn] + bm[vn])*0.5) /
    	    	  (1.0 + (am[vn] + bm[vn])*0.5); 

    if (tval>0) {
      m1tab[vn] = tval;				/* C-N implicit */
      m2tab[vn] = am[vn] / (1.0+(am[vn]+bm[vn])*0.5); 
    }
    else {					/* purely implicit */
      m1tab[vn] = 1.0    / (1.0+am[vn]+bm[vn]);
      m2tab[vn] = am[vn] / (1.0+am[vn]+bm[vn]); 
    }

/*  fprintf (stderr,"t %d %4d %10.5g %10.5g\n",
		t,vn-RATESIZ/2,am[vn],bm[vn]);  /* */

/* fprintf (stderr,"t %d %4d %10.5g %10.5g\n",
		t,vn-RATESIZ/2,m1tab[vn],m2tab[vn]);  /* */

   }  /* for (vn;;) */
 }
 chp->oldtiminci = timinc;
 chp->oldvi = -1e10;
 chp->oldmr = -1e10;
}


/*----------------------------------------*/

void makchanimpl (chanparm *chp)

/* Make alpha, beta rate tables and implicit tables for 
   one state variable (either HH type or sequential-state).
   If chp->nitab is nonzero, implicit tables will be made
   for an HH channel.
*/

{
  makchanratetab (chp);		/* make rate function tables */
  makchanimpltab (chp);		/* make implicit tables */
}

/*----------------------------------------*/

void calcchanimpl(double v, double mrate, chanparm *chp)
                       
/* Used at runtime, for HH chans only. Given input voltage in
   volts, lookup HH rate constants in tables that were constructed
   as a function of millivolts above -250 mv. Do interpolation
   between adjacent table entries and calculate m1, m2 values
   from implicit lookup tables.  Tables give second-order equation
   for HH rate constants.  Create tables if they have not been
   created yet.

*/

{
  int vn,vnn,ratesiz;
  static double x,r,rr,tval,mr;
  double a,b,ab,ab2, val;
  float *am,*bm;
  float *m1tab,*m2tab;
  double m1, m2;

  if ((!chp->chanitab) || (timinc!=chp->oldtiminci)) makchanimpl(chp);
  if (v != chp->oldvi || mrate != chp->oldmr) {

    ratesiz = RATESIZ+1;
    m1tab = chp->chanitab;
    m2tab = chp->chanitab + ratesiz;
 
    if (ncabs(v) < 1e-20) v = 0.0;
    x = (v * 1000.0) + (RATESIZ/2.0);		/* convert to mv + 250 */
    if (x < 0) x = 0;
    if (x > RATESIZ) x = RATESIZ;
    vn = (int)x;				/* table starts at -250 mv */
    vnn = vn + 1;

    r = x - vn;				/* remainder for interpolation */
    rr = 1.0 - r;

    if (mrate==1.0) {			/* find activation rates */
      m1  =  m1tab[vn]*rr + m1tab[vnn]*r;	/* interpolate m1 table */ 
      m2  =  m2tab[vn]*rr + m2tab[vnn]*r;	/* interpolate m2 table */ 
    }
    else {		/* non-standard tau */
      am = chp->chanftab;
      bm = chp->chanftab + ratesiz;

      if (mrate==0) mrate = 1e-4;
      mr = mrate;

      a = (am[vn]*rr + am[vnn]*r) * mr;
      b = (bm[vn]*rr + bm[vnn]*r) * mr;
      ab  = a+b;
      ab2 = ab*0.5;

      tval  =	  (1.0 - ab2) /
    	    	  (1.0 + ab2); 

      if (tval>0) {
        m1 = tval;				/* C-N implicit */
        m2 = a / (1.0+ab2); 
      }
      else {					/* purely implicit */
        m1 = 1.0 / (1.0+ab);
        m2 = a   / (1.0+ab); 
      }
    }
    chp->oldvi = v;
    chp->oldmr = mrate;
    chp->oldtiminci = timinc;
    chp->ival[0] = m1;				/* save values for impl calc */
    chp->ival[1] = m2;
  }
}
/*--------------------------------------------*/

void calcchanrate(double v, chanparm *chp)

/* Calculate rates from tables. Note that rate tables start at 0. */

{
  int i,vn,vnn,ratesiz;
  double x,r;
  float *am,*bm;
  float *chanftab;
  double *fval;

  if (!chp->chanftab || (timinc!=chp->oldtimincf)) makchanratetab (chp);

  if (v != chp->oldvf) {

   fval = chp->fval;
   for (i=0; i<chp->nfval; i++) {
     ratesiz = RATESIZ+1;
     chanftab = chp->chanftab + i*ratesiz;

     if (ncabs(v) < 1e-20) v = 0.0;
     x = (v * 1000.0) + (RATESIZ/2.0);		/* convert to mv + 250 */
     if (x < 0) x = 0;
     if (x > RATESIZ) x = RATESIZ;
     vn = (int)x;			     /* table starts at -250 mv */
     r = x - vn;			     /* remainder for interpolation */
 
     fval[i] = chanftab[vn];		     /* alpha, beta, etc. */
     fval[i] += (chanftab[vn+1] - fval[i]) * r; /* interpolate table */
   } 

   chp->oldvf = v;
   chp->oldtimincf = timinc;

/* fprintf (stderr,"v %g vn %d am %g bm %g\n",
		v,vn,fval[0],fval[1]); /* */
  }
}

/*--------------------------------------------*/

void chanrate(chan *cpnt, chanparm *chp)
             
/* Given chan, find its voltage in volts, lookup alpha and 
   beta rate functions in tables that were constructed as a 
   function of millivolts above -250 mv. Do interpolation 
   between adjacent table entries. Used by markov sequential-state 
   integration method.
*/

{
     double voff;

  if (!cpnt || !cpnt->comp1) return;
  voff = (chp->pn ? cpnt->voffsh: cpnt->voffsm);
  calcchanrate(cpnt->comp1->v - voff, chp);
}

/*--------------------------------------------*/

void chanimpl(chanparm *chp, chan *cpnt)
             
/* Given chan, find its voltage in volts, lookup alpha and 
   beta rate functions in tables that were constructed as a 
   function of millivolts above -250 mv. Do interpolation 
   between adjacent table entries. Used by markov sequential-state 
   integration method.
*/

{
//  if (!cpnt || !cpnt->comp1) return;
  if (chp->pn == 0) 
       calcchanimpl(cpnt->comp1->v - cpnt->voffsm, cpnt->arate, chp);
  else calcchanimpl(cpnt->comp1->v - cpnt->voffsh, cpnt->brate, chp);
}

/*----------------------------------------*/

void makchan_dftab(chantype *chtyp, int ndftab)

/* Make tables for driving force functions (e.g. GHK current equation)
   for any channel type.
   Makes "ntab" tables (e.g. am, bm, etc.).
*/

{

}

/*--------------------------------------------*/

void calcchan_dftab(double v, chantype *chtyp)

{

}

/*--------------------------------------------*/

double calcchaninf(double v, chanparm *chp)

{
  double alpham, betam;

  calcchanrate(v,chp);
  alpham = chp->fval[0];
  betam = chp->fval[1];
  return (alpham / (alpham + betam));
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double calcchantau(double v, chanparm *chp)

{
  double alpham, betam;

 calcchanrate(v,chp);
 alpham = chp->fval[0];
 betam = chp->fval[1];
 return (timinc / (alpham + betam));
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double calcchantaua(double v, chanparm *chp)

{
  double alpham;

 calcchanrate(v,chp);
 alpham = chp->fval[0];
 return (timinc / alpham);
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double calcchantaub(double v, chanparm *chp)

{
  double betam;

 calcchanrate(v,chp);
 betam = chp->fval[1];
 return (timinc / betam);
}

/*--------------------------------------------*/

double al0(chan *cpnt)

{
    chanparm *chp;

  chp = &cpnt->chtyp->parm[0];
  chanrate(cpnt, chp);
  return chp->fval[0];
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double bet0(chan *cpnt)
{
    chanparm *chp;

  chp = &cpnt->chtyp->parm[0];
  chanrate(cpnt, chp);
  return chp->fval[1];
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double al1(chan *cpnt)
{
    chanparm *chp;

  chp = &cpnt->chtyp->parm[1];
  chanrate(cpnt, chp);
  return chp->fval[0];
}

/* - - - - - - - - - - - - - - - - - - - - - -*/

double bet1(chan *cpnt)
{
    chanparm *chp;

  chp = &cpnt->chtyp->parm[1];
  chanrate(cpnt, chp);
  return chp->fval[1];
}

/*--------------------------------------------*/

double getntresp (chan *ch, comp *cpnt)

/* Calculate a channel's response to ligands present. 
   Agonist ligands have positive weight, and
   antagonist ligands have negative weight.

   Not totally correct, but simple; otherwise,
   need to have a different binding func for 
   each ligand (see "rantag()" and "ragonist()") below.
*/
   
{
   int type;
   ntcomp *npnt;
   double a,ampl;

  ampl = 0.0;
  for (npnt=cpnt->ntpnt; npnt; npnt=npnt->ntpnt) { /* search list */
	type = npnt->ctype - GLU;
        if (type <= 0) type = 0;
        if (type >= NNTR) type = NNTR;
        a = ch->chtyp->respamp[type];
        ampl += a * npnt->val; 
  }
  if (ampl < 0) ampl = 0;
  return ampl; 
}

/*--------------------------------------------*/

double getagonresp (chan *ch, comp *cpnt)

/* Calculate a channel's response to agonist ligands present */ 
/* Agonist ligands have positive weight */
   
{
   int type;
   ntcomp *npnt;
   double a,ampl;

  ampl = 0.0;
  for (npnt=cpnt->ntpnt; npnt; npnt=npnt->ntpnt) { /* search list */
	type = npnt->ctype - GLU;
        if (type <= 0) type = 0;
        if (type >= NNTR) type = NNTR;
        a = ch->chtyp->respamp[type];
	if (a > 0) ampl += a * npnt->val; 
  }
  if (ampl < 0) ampl = 0;
  return ampl; 
}

/*--------------------------------------------*/

double getantresp (chan *ch, comp *cpnt)

/* Calculate a channel's response to competitive antagonists present.  
   Antagonists have negative weight in the respamp[] table.
    Sum all the negative weights and invert.
*/
   
{
   int type;
   ntcomp *npnt;
   double a,ampl;

  ampl = 0.0;
  for (npnt=cpnt->ntpnt; npnt; npnt=npnt->ntpnt) { /* search list */
	type = npnt->ctype - GLU;
        if (type <= 0) type = 0;
        if (type >= NNTR) type = NNTR;
        a = ch->chtyp->respamp[type];
	if (a < 0) ampl -= a * npnt->val; 
  }
  if (ampl < 0) ampl = 0;
  return ampl; 
}

/*--------------------------------------------*/

int isntresp (chan *ch, comp *cpnt)

/* Find if a channel can respond to ligands present */ 
   
{
   int type;
   ntcomp *npnt;
   double a,ampl;

  ampl = 0.0;
  for (npnt=cpnt->ntpnt; npnt; npnt=npnt->ntpnt) { /* search list */
	type = npnt->ctype - GLU;
        if (type <= 0) type = 0;
        if (type >= NNTR) type = NNTR;
        a = ch->chtyp->respamp[type];
	if (a > 0) return 1;
  }
  return 0;
}

/*--------------------------------------------*/

int isantag (chan *ch, comp *cpnt)

/* Find if a channel can respond to antagonists present */ 
   
{
   int type;
   ntcomp *npnt;
   double a,ampl;

  ampl = 0.0;
  for (npnt=cpnt->ntpnt; npnt; npnt=npnt->ntpnt) { /* search list */
	type = npnt->ctype - GLU;
        if (type <= 0) type = 0;
        if (type >= NNTR) type = NNTR;
	a = ch->chtyp->respamp[type];
	if (a < 0) return (1);
  }
  return 0;
}

/*--------------------------------------------*/

double rchanf(chan *cpnt)

/* Forward binding rate constant for channel opening.
   Calculated from channel conductance.
   For use in dochan2().
 */

{
   double conduct;

 if (!cpnt) return (0.0);
 if (cpnt->comp2)
   conduct=((synap *)cpnt->comp2)->conduct;	/* find synap conductance */
 else
   conduct=cpnt->conduct;			/* find chan conductance */
 return (conduct*qrate(cpnt->chtyp->parm));	/* rate per second */
}

/*--------------------------------------------*/

double rchanr(chan *cpnt)

/* Reverse binding rate constant for channel opening.
   Calculated from channel conductance.
   For use in dochan2().
 */

{
   double conduct;

 if (!cpnt) return (0.0);
 if (cpnt->comp2)
   conduct=((synap *)cpnt->comp2)->conduct;	/* find synap conductance */
 else
   conduct=cpnt->conduct;			/* find chan conductance */
 return ((1.0-conduct)*qrate(cpnt->chtyp->parm));	/* rate per second */
}

/*--------------------------------------------*/

double rt(chan *spnt)

/* Return a constant rate scaled by time increment. */

{
  return (qrate(spnt->chtyp->parm));
}

/*--------------------------------------------*/

double rnt(chan *cpnt)

/* Forward binding rate constant for neurotransmitter.
   For use in Markov schemes where agonist and antagonist
   concentrations are weighted postive and negative,
   respectively and are summed to get one effect.
   (Not strictly very correct, but simple.)

   Return neurotrans concentration, multiplied by
   timinc, Q10 value, and scaling factor "trconc" for use in dochan2().
 */

{
   double nt;
   synap *spnt;

  if (!cpnt) return (0.0);
  if (isntresp(cpnt,cpnt->comp1)) {      /* find postsyn ligand conc */
     nt = getntresp (cpnt,cpnt->comp1);  /* if it exists, use it */
  }
  else {                                 /* otherwise use direct signal */
    if (spnt=(synap*)cpnt->comp2) {
      nt = spnt->conduct * spnt->trconc;   /* synaptic nt conc */
    }
    else nt = 0;
  }
  return (nt*qrate(cpnt->chtyp->parm));    /* rate prop. to nt conc.*/
}                                       /*  per second */

/*--------------------------------------------*/

double rca(chan *cpnt)

/* Forward binding rate constant for Ca in KCa chans.
   For use in Markov schemes where Ca gates channel.

   Return Ca concentration, multiplied by timinc and Q10 value.
 */

{
   double ca;
   cacomp *capnt;

  if (!cpnt) return (0.0);
  if (!(capnt=cpnt->comp1->capnt)) {
        ca = dcai;
  }
  else ca = capnt->cais[0];
  return (ca*qrate(cpnt->chtyp->parm));    /* rate prop. to ca conc.*/
}                                       /*  per second */

/*--------------------------------------------*/

double ragonist(chan *cpnt)

/* Forward binding rate constant for neurotransmitter.
   For use in Markov schemes where separate functions
   are used for agonist and antagonist concentrations.
 
   Return neurotrans concentration, multiplied by
   timinc and scaling factor "trconc" for use in dochan2().
 */

{
   double nt;
   synap *spnt;

  if (!cpnt) return (0.0);
  if (isntresp(cpnt,cpnt->comp1)) {      /* find postsyn ligand conc */
     nt = getagonresp (cpnt,cpnt->comp1);  /* if it exists, use it */
  }
  else {                                 /* otherwise use direct signal */
    if (spnt=(synap*)cpnt->comp2) {
      nt = spnt->conduct * spnt->trconc;   /* synaptic nt conc */
    }
    else nt = 0;
  }
  return (nt*qrate(cpnt->chtyp->parm));    /* rate prop. to nt conc.*/
}                                       /*  per second */

/*--------------------------------------------*/

double rantag(chan *cpnt)

/* Forward binding rate constant for competitive antagonist 
   ligands.  Return neurotrans concentration, multiplied by
   timinc and scaling factor "trconc" for use in dochan2().
 */

{
   double nt;
   synap *spnt;

  if (!cpnt) return (0.0);
  if (isantag(cpnt,cpnt->comp1)) {      /* find postsyn ligand conc */
     nt = getantresp (cpnt,cpnt->comp1);  /* if it exists, use it */
     return (nt*qrate(cpnt->chtyp->parm));    /* rate prop. to nt conc.*/
  }
  else return 0;
}                                       /*  per second */

/*--------------------------------------------*/

void mak2state(chanstate *spnt, double (*frate)(chan *cpnt),
				double (*rrate)(chan *cpnt), double rate)

/* number states from 0 to n; 
   set numstate = n;
   set numtrans = number of transitions.
   set cond     = conductance of state;

   for each state, set transitions:
    set trans = state to go to on this transition. 
    set trate = function that returns basic rate for transition. 
    set ratemul = multiplier for rate function.
    set rateo  =  Set to 0 for m, set to 1 for h:  sets voffset and tau.
*/

  {
     double a,b;

   if (!spnt) {
	fprintf (stderr,"mak2state: no chanstate\n");
	return;
   }

   a = rate;
   b = rate;
					/*      states     */

   					/*    0   <->  1   */ 
   spnt[0].numtrans   = 1;
   spnt[0].cond       = 0;
   spnt[0].trans  [0] = 1;
   spnt[0].trate  [0] = frate;
   spnt[0].ratemul[0] = a;
   spnt[0].rateo  [0] = 3;		/* use drate = for noise */

   spnt[1].numtrans   = 1;
   spnt[1].cond       = 1.0;		/* 1 = the open state */
   spnt[1].trans  [0] = 0;
   spnt[1].trate  [0] = rrate;
   spnt[1].ratemul[0] = b;
   spnt[1].rateo  [0] = 3;
}

