/* ncelem.cc */

/* Routines to make and access "neural elements" */

#include "nc.h"
#include "y.tab.h"
#include "ncsub.h"
#include "ncelem.h"

#define EHASHSIZ 509             /* prime number to make better hash */

elem *ehashtab[EHASHSIZ] = {0};  /* element hash table */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>

#ifdef __cplusplus
}
#endif

extern elem *elempnt;
extern elem *elemend;
extern int cumelem;
extern int delelmt;
extern int cumattr;
extern int cumcattr;
extern int cumnattr;

extern elem *elpnt;

char *emalloc(unsigned int n);
void execerror(char *s, char *t);
void checkelemnode(elem *elpnt);
void checkelemnode(elem *elpnt);
float *copyfiltarr (int siz, float *oarr);
void copyattr(attrib *src, attrib *dest);
void efree(void *ptr);
chattrib *chanattr(elem *elpnt, short int ctype);

/*---------------------------------------------------*/

/*
void xdebugf(void)

{
   fprintf (stderr,"ehashtab %d\n",ehashtab[5]);
}
*/

/*---------------------------------------------------*/

void einithash(void)
{
   int i;

   for (i=0; i<EHASHSIZ; i++) 
     ehashtab[i] = (elem*)NULL;
}

/*---------------------------------------------------*/

int elemhash(int num)

/* make disordered index of element num */

{
   int i;

   i = num;
   if (i<0) i = -i;
   return (i % EHASHSIZ); 
}

/*---------------------------------------------------*/

void einstall (elem *newelem)
              
/* Install an element in hash table */
/*  Uses "direct chaining" with field "hnext". */

{
    int i;
    elem *epnt,*elast;
    static int einitfl=0;

   if (!newelem) return;

   if (!einitfl) {			/* initialize table once at start */
        einitfl = 1;
        einithash();
   }
   newelem->hnext = 0;
   i=elemhash(newelem->elnum); 		/* initial index into ehashtab*/ 
   if (!(epnt=ehashtab[i])) {            /* install directly in table */
      ehashtab[i] = newelem;
   }
   else {                               /* otherwise, go to end of list */
      for (; epnt; epnt=epnt->hnext)    /* find last element */
          elast = epnt;    
      elast->hnext = newelem;            /* put new symbol at end of list */
   }
}

/*---------------------------------------------------*/

elem *findelem(int num)
                      
/* Find element among list of all elements.
   Elements are placed in hash table, which provides
   faster access than possible with one sequential list.
   The element itself is not moved, therefore only the 
   pointers of the element list need rearranging.
   Pointers in other lists that point to element remain correct.
*/

{
   elem *epnt;
   int i,found;

        
  i = elemhash(num);
  for (found=0,epnt = ehashtab[i]; epnt; epnt = epnt->hnext) {
    if (epnt->elnum==num) {
       found = 1;
       break;
    }
  }

  if (found) return epnt;
  else {
     fprintf (stderr,"findelem: can't find elem %d\n",num);
     return (elem *)NULL; 
  }
}

/*---------------------------------------------------*/

void delhashelem (elem *pnt)

/* delete an element from the hash table */

{
   elem *epnt,*elast;
   int i,found;

  if (!pnt) return;
  
  i = elemhash(pnt->elnum);
  elast = (elem*)NULL; 
  for (found=0,epnt = ehashtab[i]; epnt; elast=epnt, epnt=epnt->hnext) {
    if (epnt->elnum==pnt->elnum) {
       found = 1;
       break;
    }
  }

  if (found) {
     if (elast) elast->hnext = epnt->hnext;	/* delete hash pointer */
     else ehashtab[i] = epnt->hnext;
  }
}

/*------------------------------------------------*/

int elemsiz(elem *epnt)

/* Return the size of a neural element in bytes. */

{
    int esize;

#ifdef DEBUG 
  if (debug & 1 && debugz & 16)  fprintf (stderr,"elemsiz %d\n",cumelem);
#endif

  if (!epnt) return(0);

  switch (epnt->ctype) {

        default:      esize = 0;		break;
 	case ELEMENT: esize = sizeof(elem);	break;
 	case SPHERE:  esize = sizeof(sphere);	break;
 	case CABLE:   esize = sizeof(cable);	break;
 	case SYNAPSE: esize = sizeof(synapse);	break;
 	case ROD:     esize = sizeof(photrec);	break;
 	case CONE:    esize = sizeof(photrec);	break;
 	case TRANSDUCER: esize = sizeof(photrec); break;
 	case GJ:      esize = sizeof(gapjunc);	break;
 	case RESISTOR:esize = sizeof(resistor);	break;
 	case LOAD:    esize = sizeof(loadelem);	break;
 	case GNDCAP:
 	case CAP:     esize = sizeof(capac);	break;
 	case BATT:    esize = sizeof(batt);	break;
 	case BUF:     esize = sizeof(vbuf);	break;
 	case CHAN:    esize = sizeof(elem);	break;
  }
  return esize;
}

/*------------------------------------------------*/

elem *tmpelem(void)

/* Set up an element to hold
   nodes and locations temporarily. */

{
  static elem *epnt, nullelem = {0};

  epnt = &nullelem;
  epnt->node1a = NULLVAL;	/* reset node number */ 
  epnt->node1b = NULLVAL;
  epnt->node1c = NULLVAL;
  epnt->node1d = NULLVAL;
  epnt->node2a = NULLVAL;
  epnt->node2b = NULLVAL;
  epnt->node2c = NULLVAL;
  epnt->node2d = NULLVAL;
  epnt->attpnt = (attrib *)NULL;
  epnt->nodp1  = (node *)NULL;
  epnt->nodp2  = (node *)NULL;
  epnt->next   = (elem *)NULL;
  epnt->last   = (elem *)NULL;
  epnt->hnext  = (elem *)NULL;
  epnt->lptr   = (elem *)NULL;
  epnt->saved  = 0;
  epnt->jnoise = 0;
  epnt->rsd    = 0;
  epnt->modif  = 0;

  epnt->ctype = ELEMENT;	/* element is default type */
  elpnt = (elem *)NULL;		/* reset current element pointer */
  return (epnt);
}

/*------------------------------------------*/

elem *elemcopy (elem *epnt, char *(*dalloc)(unsigned int))

/* Copy a neural element to a new location using dalloc to
    allocate space. */

{
    int i,siz;
    char *src,*dest;

  siz = elemsiz(epnt);
  src = (char*)epnt;
  if (!(dest=dalloc(siz))) {
    fprintf (stderr,"elemcopy, not enough space.\n");
    return (elem*)NULL;
  }
  for (i=0; i<siz; i++) {
    dest[i] = src[i];
  }
  return ((elem*)dest);
}

/*------------------------------------------*/

elem *savelem (elem *epnt)

/* Copy an element to the "save heap" */
/*  and free its old space. */

{
      elem *nepnt;
      attrib *apnt, *napnt, *onapnt, *tapnt;

    if (epnt->saved > 1) return epnt;			/* don't free again */
    nepnt = elemcopy(epnt,emalloc);			/* copy the element */
    if (!nepnt) {
      fprintf (stderr,"Can't save element.\n");
      return (elem *)NULL;
    }
    for (onapnt=(attrib*)NULL,apnt=epnt->attpnt; apnt; ) { /* copy attribs */ 
         int i,siz;
         char *src, *dest;
      switch (apnt->ctype) {
        case CHNOISE:
        case VESNOISE: siz = sizeof(nattrib);  break;
        case CGMP:
        case NMDA:
        case CACOMP:   siz = sizeof(cattrib);  break;
         default:      siz = sizeof(chattrib); break;
      }
      if (!(napnt=(attrib*)emalloc(siz))) {     /* allocate new attrib space */
         fprintf (stderr,"savelem: not enough space for new attrib.\n");
         return (elem*)NULL;
      }
      src = (char *)apnt;
      dest = (char *)napnt;
      for (i=0; i<siz; i++) {
	 dest[i] = src[i];
      }
      napnt->attpnt = (attrib *)NULL;
      if (onapnt) onapnt->attpnt = napnt;     /* set pointer */
      else         nepnt->attpnt = napnt;
      tapnt = apnt;
      onapnt = napnt;
      apnt = apnt->attpnt;
      efree (tapnt);                          /* free old attrib */

    }
    if (elempnt==epnt) elempnt = nepnt;		/* set pointers */
    if (elemend==epnt) elemend = nepnt;
    if (epnt->last) epnt->last->next = nepnt;
    if (epnt->next) epnt->next->last = nepnt;
    einstall(nepnt);
    delhashelem(epnt); 
    switch (epnt->ctype) {			/* copy arrays */
	  int nfilt;
          synapse *spnt, *nspnt;

     case SYNAPSE: spnt = (synapse*)epnt;
	           nspnt = (synapse*)nepnt;

		if ((nfilt=spnt->nfilt1)!=NULLVAL) 
		  nspnt->timec1 = copyfiltarr(nfilt, spnt->timec1);
     		if ((nfilt=spnt->nfilt1h)!=NULLVAL)
		  nspnt->timec1h = copyfiltarr(nfilt, spnt->timec1h);
     		if ((nfilt=spnt->nfilt2)!=NULLVAL) 
		  nspnt->timec2 = copyfiltarr(nfilt, spnt->timec2);
     		if ((nfilt=spnt->nfilt3)!=NULLVAL) 
		  nspnt->timec3 = copyfiltarr(nfilt, spnt->timec3);
		 break;
    }
    efree (epnt); 
    if (nepnt->saved==1) nepnt->saved++;	/* don't free again */
    return (nepnt);
}

/*------------------------------------------*/

void elcopy (elem *src, elem *dest)

/* Copy a neural element's node information. */

{
  if (!src || !dest) return;

  dest->ctype  = src->ctype;
  dest->node1a = src->node1a;
  dest->node1b = src->node1b;
  dest->node1c = src->node1c;
  dest->node1d = src->node1d;
  dest->node2a = src->node2a;
  dest->node2b = src->node2b;
  dest->node2c = src->node2c;
  dest->node2d = src->node2d;
  dest->attpnt = src->attpnt;
  dest->nodp1 = src->nodp1;
  dest->nodp2 = src->nodp2;
  dest->next = src->next;
  dest->last = src->last;
  dest->hnext = src->hnext;
  dest->lptr  = src->lptr;
  dest->saved = src->saved;
  dest->jnoise = src->jnoise;
  dest->rsd   = src->rsd;
  dest->modif = src->modif;
  dest->elnum = src->elnum;
}

/*------------------------------------------------*/

elem *makelem(int etype, elem *oepnt)

/* Make a new neural element, and link it
   to the elem list. Return a pointer to the 
   new element. Before doing this, though,
   check the old element (oepnt).  If it exists
   and has been already been initialized, then
   don't make a new element, but return the old
   one. */

{
    elem *epnt;
    int esize;

#ifdef DEBUG 
  if (debug & 1 && debugz & 16)  fprintf (stderr,"elem %d\n",cumelem);
#endif

  if (oepnt && oepnt->ctype!=ELEMENT) {	/* If old element exists, */
        checkelemnode(oepnt);		/*  set its node pointers */
	return (oepnt);			/*  and return it. */ 
  }

  switch (etype) {

        default:
 	case ELEMENT: esize = sizeof(elem);	break;
 	case SPHERE:  esize = sizeof(sphere);	break;
 	case CABLE:   esize = sizeof(cable);	break;
 	case SYNAPSE: esize = sizeof(synapse);	break;
 	case ROD:     esize = sizeof(photrec);	break;
 	case CONE:    esize = sizeof(photrec);	break;
 	case TRANSDUCER: esize = sizeof(photrec); break;
 	case GJ:      esize = sizeof(gapjunc);	break;
 	case RESISTOR:esize = sizeof(resistor);	break;
 	case LOAD:    esize = sizeof(loadelem);	break;
 	case GNDCAP:
 	case CAP:     esize = sizeof(capac);	break;
 	case BATT:    esize = sizeof(batt);	break;
 	case BUF:     esize = sizeof(vbuf);	break;
 	case CHAN:    esize = sizeof(elem);	break;
  }

#ifndef XSTIMM
  if ((epnt=(elem *)emalloc(esize)) == NULL) {
     fprintf (stderr,"no space left for elem %d\n",cumelem);
     return ((elem*)NULL);  
  }

#else		/* if we are ignoring neural elements in "stim" */

    {  static char tmpelement[sizeof(synapse)*2] = {0};

       epnt = (elem *)&tmpelement;
    }
#endif

  if (oepnt) {
    elcopy (oepnt, (elem *)epnt);	/* first, copy node info. */
    checkelemnode(epnt);		/*  set elem's node pointers */
  }					/*  and the nodes' elem pointers */
 
  else {				/* else reset */
    epnt->saved = 0;
    epnt->modif = 0;
  }

  epnt->next = (elem *)NULL;		/* next, set pointers */
  if (!elempnt) elempnt = epnt;  	/* save head if first branch */
  if (elemend) {
    elemend->next = epnt;
    epnt->last = elemend;
  }
  else  {
    epnt->last = (elem *)NULL;
  }
  elemend = epnt;

  epnt->lptr  = (elem *)NULL;
  epnt->ctype  = etype;
  epnt->elnum = ++cumelem; 			/* increment total */
  einstall (epnt);				/* install in hash table */

  switch (etype) {

        default:
 	case ELEMENT: break;

 	case SPHERE:	((sphere *)epnt)->dia    = NULLVAL;
			((sphere *)epnt)->Rm     = NULLVAL;
		 	((sphere *)epnt)->Cm     = NULLVAL;
		 	((sphere *)epnt)->vrev   = NULLVAL;
		 	((sphere *)epnt)->vrest  = NULLVAL;
		 	break;

 	case CABLE:	((cable *)epnt)->dia    = NULLVAL;
		 	((cable *)epnt)->dia2   = NULLVAL;
		 	((cable *)epnt)->Ri     = NULLVAL;
		 	((cable *)epnt)->Rm     = NULLVAL;
		 	((cable *)epnt)->Cm     = NULLVAL;
		 	((cable *)epnt)->vrev   = NULLVAL;
		 	((cable *)epnt)->vrest  = NULLVAL;
		 	((cable *)epnt)->length = NULLVAL;
		 	((cable *)epnt)->cplam  = NULLVAL;
		 	break;

 	case SYNAPSE:	((synapse *)epnt)->thresh = NULLVAL;
		 	((synapse *)epnt)->timec1 = (float *)NULL;
		 	((synapse *)epnt)->nfilt1 = NULLVAL;
		 	((synapse *)epnt)->timec1h = (float *)NULL;
		 	((synapse *)epnt)->nfilt1h = NULLVAL;
		 	((synapse *)epnt)->filt1hg = NULLVAL;
		 	((synapse *)epnt)->timec2 = (float *)NULL;
		 	((synapse *)epnt)->nfilt2 = NULLVAL;
		 	((synapse *)epnt)->timec3 = (float *)NULL;
		 	((synapse *)epnt)->nfilt3 = NULLVAL;
		 	((synapse *)epnt)->tfall2 = NULLVAL;
		 	((synapse *)epnt)->tfall3 = NULLVAL;
		 	((synapse *)epnt)->cgain  = NULLVAL;
		 	((synapse *)epnt)->maxcond= NULLVAL;
		 	((synapse *)epnt)->trconc = NULLVAL;
		 	((synapse *)epnt)->ntact  = NULLVAL;
		 	((synapse *)epnt)->nkd    = NULLVAL;
		 	((synapse *)epnt)->npow   = NULLVAL;
		 	((synapse *)epnt)->ngain  = NULLVAL;
		 	((synapse *)epnt)->ckd    = NULLVAL;
		 	((synapse *)epnt)->cgain  = NULLVAL;
		 	((synapse *)epnt)->curve  = NULLVAL;
		 	((synapse *)epnt)->mesng1 = NULLVAL;
		 	((synapse *)epnt)->mesng2 = NULLVAL;
		 break;

	case TRANSDUCER:
	case CONE:
 	case ROD:	((photrec *)epnt)->xpos   = NULLVAL;
		 	((photrec *)epnt)->ypos   = NULLVAL;
		 	((photrec *)epnt)->zpos   = NULLVAL;
		 	((photrec *)epnt)->dia    = NULLVAL;
		 	((photrec *)epnt)->maxcond= NULLVAL;
		 	((photrec *)epnt)->pigm   = NULLVAL; 
		 	((photrec *)epnt)->pathl  = NULLVAL;
		 	((photrec *)epnt)->attf   = NULLVAL;
		 	((photrec *)epnt)->timec1 = NULLVAL;
		 	((photrec *)epnt)->filt   = NULLVAL;
		 	((photrec *)epnt)->qc     = NULLVAL;
		 	((photrec *)epnt)->qcond  = NULLVAL;
		 	((photrec *)epnt)->save   = 0;
		 	((photrec *)epnt)->restore= 0;
		 	((photrec *)epnt)->darknoise=NULLVAL;
		 	((photrec *)epnt)->photnoise=NULLVAL;
		 	((photrec *)epnt)->prseed=0;
		 	break;

 	case GJ:      	((gapjunc *)epnt)->area   = NULLVAL;
		 	((gapjunc *)epnt)->specres= NULLVAL;
		 	((gapjunc *)epnt)->vgain  = NULLVAL;
		 	((gapjunc *)epnt)->voff   = NULLVAL;
		 	((gapjunc *)epnt)->taun   = NULLVAL;
		 	((gapjunc *)epnt)->rvgain = NULLVAL;
		 	((gapjunc *)epnt)->rvoff  = NULLVAL;
		 	((gapjunc *)epnt)->rtaun  = NULLVAL;
		 	((gapjunc *)epnt)->gmax   = NULLVAL;
		 	((gapjunc *)epnt)->gnv    = NULLVAL;
		 	((gapjunc *)epnt)->nodeca = NULLVAL;
		 	((gapjunc *)epnt)->nodecb = NULLVAL;
		 	((gapjunc *)epnt)->nodecc = NULLVAL;
		 	((gapjunc *)epnt)->nodecd = NULLVAL;
		 	((gapjunc *)epnt)->modtyp = 0;
		 	((gapjunc *)epnt)->sign   = 0;
		 	break;

 	case RESISTOR:	((resistor *)epnt)->z = NULLVAL;
		 	break;

 	case LOAD:	 ((loadelem *)epnt)->z = NULLVAL;
		 	((loadelem *)epnt)->vrev  = NULLVAL;
		 	((loadelem *)epnt)->vrest = NULLVAL;
		 	break;

 	case GNDCAP:
 	case CAP:	((capac *)epnt)->c  = NULLVAL;
			((capac *)epnt)->vrest = NULLVAL;
		 	break;

 	case BATT:	((batt *)epnt)->v  = NULLVAL;
		 	break;

 	case BUF:	((vbuf *)epnt)->delay = NULLVAL;
		 	break;

  }	/* switch (etype)  to initialize the elements */

  return (epnt); 
}

/*---------------------------------------------------*/

void delelem (elem *epnt)

/* Delete an element, including its attributes and 
    its hash table entry, and list pointers but not
    any node pointers to it.
*/

{
    attrib *apnt, *tapnt;
    elem *last,*next;

    for (apnt=epnt->attpnt; apnt; ) {   /* free element's attributes first */
        tapnt = apnt;
        switch (apnt->ctype) {
	  case CHNOISE:
	  case VESNOISE: cumnattr--; break;
	  case CGMP:
	  case NMDA:
	  case CA:
	  case CACOMP:   cumcattr--; break;
              default:   cumattr--;  break;
        }
        apnt = apnt->attpnt;
        efree (tapnt);
    }
    next = epnt->next;
    last = epnt->last;
    if (last) last->next = next;
    if (next) next->last = last;
    if (elempnt==epnt)			/* is elempnt the one being erased? */
        elempnt = next;			/*  yes, then incr elempnt */
    if (elemend==epnt)			/* is elemend the one being erased? */
        elemend = last;			/*  yes, then decr elemend */
    delhashelem(epnt); 
    switch (epnt->ctype) {
	float *filt;

     case SYNAPSE: if (filt=((synapse*)epnt)->timec1) efree (filt);
     		   if (filt=((synapse*)epnt)->timec1h)efree (filt);
     		   if (filt=((synapse*)epnt)->timec2) efree (filt);
     		   if (filt=((synapse*)epnt)->timec3) efree (filt);
		  break;
	default:	break;
    }
    efree (epnt);
    delelmt++;
/*    cumelem--;  */			/* don't decrement elem number */
}


