/* Module ncsub in Program nc */

/* Models neuronal circuits */

/* Simulates a neuron with a compartmental model.
It assumes a dendrite is made up of small isopotential
compartments which are modeled as electrical nodes.
Each node is connected to its neighbors through axial 
resistances, and has conduction pathways to ground
through membrane resistance and capacitance.
Additional conduction pathways are allowed for any node. 

Unless otherwise noted, all length values are calibrated
in meters, all voltages are in volts, all currents are
in amperes, and capacitance is in farads.

        Oct 87                  R.G. Smith

*/

#include "adef.h"
#include "nc.h"
#include "y.tab.h"
#include "control.h"
#include "ncsub.h"
#include "ncomp.h"
#include "ncplot.h"
#include "ncelem.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#ifdef CPML
#include <cpml.h>
#else
#ifdef CPML
#include <cpml.h>
#else
#include <math.h>
#endif
#endif
#include "gr.h"
double floor(double);

#ifdef __cplusplus
}
#endif

#define DEBUG

/*---------------------------------------*/

/* From "nclist.h": */

extern int cumelem;
extern int cumattr;
extern int cumnattr;
extern int cumcattr;
extern int cumnode;
extern int cumcomp;
extern int cumconn;
extern int cumchan;
extern int cumsynap;
extern int cumload;
extern int cumrec;
extern int cumrecst;
extern int cumcacomp;
extern int cumcycacomp;
extern int cumcycgcomp;
extern int cumclst;
extern int cumdlst;
extern int cumlpfilt;
extern int cumrand;
extern int cumntcomp;
extern int cumgj;
extern int reccum;
extern int delelmt;

extern comp *compnt;
extern comp *compend;
extern conn *connpnt;
extern conn *connend;
extern synap *synpnt;
extern synap *synend;
extern chan *chanpnt;
extern chan *chanend;
extern load *loadpnt;
extern load *loadend;
extern recep *recpnt;
extern recep *recend;
extern recstim *recspnt;
extern recstim *recsend;

int ztotsiz=0;

/*---------------------------------------*/

extern int setdisp;
extern int setprmap;
extern int disp_ray;
extern double setlamcrit;
extern int pinitfl;			/* =1 -> plotinit() run */
extern synap *synpnt;
extern comp *compnt;
extern recep *recpnt;
extern recstim *recspnt;
extern recstim *recsend;
extern node *nodepnt;
extern elem *elempnt;
extern elem *elemend;

extern Symbol *timeptr;
extern Symbol *plotiptr;
extern Symbol *ncompptr;

int iflag;                              /* interrupt flag set by onintr()  */

plotfr plotnod[PLOTNODSIZ] = {0};       /* node numbers, vmin, vmax voltages */

int numplots = 0;                       /* number of plots displayed */
int runyet=0;
int nocond = 0;				/* no condense */

double record(int cnod1, int cnod2, int cnod3, int cnod4, int pmod, int pval);
double getcurv(synap *spnt, double val);
double rsens(recep *rpnt, double wavel, int filt);
double ncabs(double x);
double lkexp(double val);
void initcomp(int sav_elem);
void condense();
void prcomp();
void chksmall();
void compcon();
void runmod(double timestep, int run);
void plotinit(int numplots);
void runstim(double xtime, int nsteps, int endfl);
void runsyn(int nsteps);
void runrec(int t); 
double getrecval(int i);
void mplot(double val, double time, int numplots, int i, int pflag);
void runcomp();
node *findnode(nodeint num1, nodeint num2, 
			nodeint num3, nodeint num4, char *str);
int findrecep (int cnod1, int cnod2, int cnod3, int cnod4,
			 recep **rpnt, char *s);
void execerror (char *s, char *t);
void getarg(int narg, double **val, short **vtyp);   
void rchan(synap *sdyad, int stime, double chanligand);
int poisdev(double m);
double gamdev(int a);
int readstim(double rtime);
void delrstim(recstim *rspnt);
conn *printconns(comp *pnt, int nprint);
comp *othercomp (comp *pnt,conn *cpnt);
char *prnode (int n1, int n2, int n3, int n4); 
char *rsetstate (char *state);
void restorstate();
elem *findelem (int num);
double gjalpha (double v, gj *gjpnt);
double gjbeta  (double v, gj *gjpnt);
double getcyca (gj *gjpnt);
double getcycg (gj *gjpnt);
double dochani (chan *cpnt, double critc);
recstim *getstimp1 (double stime);
recstim *getstimp2 (double stime);
double sqrt(double val);
double pow(double x, double expon);
double callproc(Symbol *procp, int npar, double par1, double par2);
double getnt (comp *cpnt, int type);
int setnt(comp *cpnt, int type, double val);
int isnt (comp *cpnt, int type);
double getmesng(gj *gjpnt, int mesng);
char *findsym(int num);
chantype *getchantype(int ctype, int cnum);
double ghkv (chantype *chtyp,double cao, double cai);
int smaxmin(double val);
int elemsiz(elem *epnt);
double callfunc(Symbol *funcp, int npar, double par1, double par2,
					 double par3, double par4);
void plotpen(int pen, int nplot);
void initrec();
 
/*------------------------------------*/

void actrun (double timestep, int run)
{
    int i;
    static int ovidmode=0;
    static elem *oelemend = 0;
    double totspace,elemspace;
    elem *epnt;

  if (setdisp>=0) disp = setdisp;	/* "-d n" overrides "disp=n" */
  if (setprmap>=0) prmap = setprmap;	/* "-p n" overrides "prmap=n" */
  if (setlamcrit>=0) lamcrit=setlamcrit; /* "-l n" overrides "lamcrit=n" */

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"actrun start %g run %d\n",timestep, run);
  if (debug & 256) nocond = 1;
#endif

    if (elemend != oelemend) {	/* if any new neural elements */
      if (!runyet) {
        if (!vidmode && !disp_ray) {
		int i;
						  /* Calculate space used */
						  /* by elements */
        for (elemspace=i=0,epnt=elempnt; epnt; epnt=epnt->next,i++) {  
          /* fprintf (stderr,"%d type %s elemsiz %d\n", 
			i,findsym(epnt->ctype),elemsiz(epnt)); */
          elemspace += elemsiz(epnt);
        }
        totspace = elemspace + cumnode * sizeof(node) + 
		cumclst * sizeof(conlst) + cumattr * sizeof(chattrib) + 
		cumcattr * sizeof (cattrib) + cumnattr * sizeof(nattrib); 
	totspace /= MB;
	if     (cumelem> 1) printf ("#c %d neural elements (%.3g MB)",
					cumelem,totspace);
	else if (cumelem==1) printf ("#c %d neural element (%.3g MB)",
					cumelem,totspace);
          printf (" converted to compartments.\n");
       }
      }
      initcomp(0);		/* make compartments */
      if (!nocond) condense();
      chksmall();		/* check for small compartments */
      compcon();		/* set up time interval in compartment */
      ncompptr->val = cumcomp;	/* set number of comps for user */
      if (!runyet) {
        if (!vidmode && !disp_ray) {
	     int remelem;

          if (ovidmode) ctext();
	  remelem = cumelem - delelmt;
	  if      (remelem> 1) printf ("#c %d neural elements saved.\n",remelem);
          else if (remelem==1) printf ("#c %d neural element saved.\n",remelem);
	  elemspace = 0;
	  if (remelem) {
             for (i=0,epnt=elempnt; epnt; epnt=epnt->next,i++) {  
               /* fprintf (stderr,"%d type %s elemsiz %d\n", 
			i,findsym(epnt->ctype),elemsiz(epnt)); */
                elemspace += elemsiz(epnt);
             }
	  }
          totspace = ((double)cumcomp*sizeof(comp)) + cumnode * sizeof(node) + 
		      cumconn * sizeof(conn) + cumrand * sizeof(randstate) +
		      cumchan * (sizeof(chan) + 4*sizeof(stconc))  +
		      cumntcomp * sizeof(ntcomp) + cumload * sizeof(load) + 
		      cumrec  * sizeof(recep) + cumrecst * sizeof(recstim) +
		      cumclst * sizeof(conlst) + cumdlst * sizeof(dyadlst) +
		      cumsynap * sizeof(synap) + 
		      cumlpfilt * (sizeof(lpfilt)+2*5*sizeof(float)) +
		      cumcacomp * (sizeof(cacomp)+ 11*sizeof(double)) +
		      cumgj * sizeof(gj) +
		      elemspace ;
	  totspace /= 1048576.0;
	  printf ("#c Total memory space used %.3g MB.\n",totspace);
	  if      (cumcomp> 1) printf ("#c %d comps",cumcomp);
          else if (cumcomp==1) printf ("#c %d comp",cumcomp);
	  if      (cumnode> 1) printf (", %d nodes",cumnode);
          else if (cumnode==1) printf (", %d node",cumnode);
	  if      (cumconn> 1) printf (", %d connections",cumconn);
          else if (cumconn==1) printf (", %d connection",cumconn);
	  if      (cumchan> 1) printf (", %d channels",cumchan);
          else if (cumchan==1) printf (", %d channel",cumchan);
	  if      (cumsynap> 1) printf (", %d synapses",cumsynap);
          else if (cumsynap==1) printf (", %d synapse",cumsynap);
	  if      (cumrec> 1) printf (", %d photorcprs",cumrec);
          else if (cumrec==1) printf (", %d photorcpr",cumrec);
          printf (".\n");
	  fflush (stdout);
        }
        if (prmap & (PCOND|PCOMP)) prcomp();
       }	/* runyet = 0 */

      else {         	/* runyet == 1 */
	if (vidmode) {
	    if (!ovidmode && !disp_ray) cgraphics(); 
	}
        else 
	    if (ovidmode) ctext();
	}
     }
  runmod(timestep,run); /* run comps, synapses and receps */
  ovidmode = vidmode;
  oelemend = elemend;
}
 
/*------------------------------------*/

void chksmall(void)

/* Check for possible errors in constructing compartments. */

{
   comp *pnt,*pnt2;
   node *npnt,*onpnt;
   int found;
   conlst *lpnt;
   conn *cpnt;

  for (pnt=compnt; pnt; pnt=pnt->next) {
    if (pnt->rm > (LARGERES * .9) || pnt->cap < SMALLCAP * 1.1) {
       fprintf(stderr,"# nc: Warning: possible compartment build error.\n");
       printf("# Warning: possible error:\n");
       printf("# Compartment # %u is missing: ",pnt->num);
       printf(" gm %6.3g, cap %6.3g",pnt->rm,pnt->cap);
       for (found=0,lpnt=pnt->nodlst; lpnt; lpnt=lpnt->next) {
            if (npnt=(node*)lpnt->conpnt) {
                printf (" at node %s.\n",
			prnode(npnt->nodenm1,npnt->nodenm2,
				npnt->nodenm3,npnt->nodenm4));
                 found = 1;
            }
       }
       if (!found) 
         for (found=0,npnt=nodepnt; npnt; npnt=npnt->next) {    /* find node */
           if (npnt->comptr == pnt) { 
                printf (" at node %s.\n",
			prnode(npnt->nodenm1,npnt->nodenm2,
				npnt->nodenm3,npnt->nodenm4));
	       found = 1;
               break; 
           }
       }
      if (found) 
         onpnt = npnt;
      else {
         printf ("# Small comp without node. ");
         onpnt = (node *)NULL;
      }
      printf ("# Check: \"");
      if (cpnt=printconns(pnt,1)) {
        pnt2 = othercomp (pnt,cpnt);		/* get other compartment */
        if (pnt2) {				/*  find its nodes */
          printf ("\" which connects");
           for (found=0,lpnt=pnt2->nodlst; lpnt; lpnt=lpnt->next) {
            if (npnt=(node*)lpnt->conpnt) {
                printf (" to node %s.\n",
			prnode(npnt->nodenm1,npnt->nodenm2,
				npnt->nodenm3,npnt->nodenm4));
                 found = 1;
            }
         }
        }  /* if pnt2 */
       }
       else printf ("\"\n");
       if (onpnt) printf ("# Node %s",
			prnode(onpnt->nodenm1,onpnt->nodenm2,
			onpnt->nodenm3,onpnt->nodenm4));
       else printf ("# It"); 
       printf (" should also connect to a cable, sphere, or load.\n");
       fflush (stdout);

    }  /*  if (pnt-rm...) */
  }  /* for (pnt=compnt;; ) */
}

/*------------------------------------*/

static Symbol *onplpnt= (Symbol *)NULL;

void onplot(void)
{
   double d=0;

   if (onplpnt) {
       timeptr->val = xtime;               /* save correct time for user */
       callproc (onplpnt,0,d,d);
       xtime = timeptr->val; 		   /* reset time */
       ploti = plotiptr->val;              /* reset ploti (in case changed) */
   }
}

/*------------------------------------*/


void setonplot(void) 

{
    Symbol *f;

  if (f=lookup("onplot")) onplpnt = f;
  else                    onplpnt = (Symbol*)NULL;
}


/*------------------------------------*/

void runmod(double timestep, int runfl)
                    
/* Run the simulation; 
   schedule the receptors every msec,
   schedule the synapses to run every 100 microsec,
   and run the compartments every "timestep".
   Plot the nodes specified using "record".
*/

{
  static double ptime, oldtime=-1e6;
  static double stoptime;
  static double time5, otime5, time4, otime4;
  static int nsteps;
  extern double endexp;
  int i,newtim;

#ifdef DEBUG
 if (debug & 1) fprintf (stderr,"runmod step %g runfl %d runyet %d\n",
			timestep,runfl,runyet);
#endif

#define ROUNDUP (stiminc * 0.001)
#define STIME5 (-1000)

 newtim = (xtime!=oldtime);
 oldtime = xtime;		/* initial time setting from user */
 setonplot();			/* try to find user-defined plot routine */
 if (numplots && !pinitfl)  plotinit(numplots);
 if (newtim) {
      ptime = ploti;
      otime4 = otime5 = STIME5;
 }
 if (runfl) {          /* run mode: run until endexp */
   stoptime = endexp - 1e-10;
 }
 else {                          /* step mode */
   stoptime = oldtime + timestep - 1e-10;
/*   if (stoptime > endexp) stoptime = endexp; */
 }
 
 if ((timinc / stiminc) > MAXLONG) {
	timinc = (MAXLONG + STIME5) * stiminc;
	printf ("# runmod: 'timinc' is too large, limited to %9.5g\n",timinc);
 }
 
 for (xtime=oldtime; xtime<stoptime; xtime+=timinc, ptime+=timinc)
  {
    if (debug & 1) fprintf (stderr,"runmod: time %g\n",xtime);

    time5 = floor (xtime / stiminc + ROUNDUP);
/* fprintf (stderr,"time5 %g otime5 %g time %30.30g\n",time5,otime5,xtime); */
    if (time5 > otime5) {
        nsteps = (int)(time5 - otime5);	/* calculate steps actually taken */
        otime5 = time5;
	nsteps = 1;
        runstim(xtime,nsteps,0);      	/* run stimuli */
        runsyn(nsteps);			/* run synapses  */
        runrec(nsteps);                 /* run receptors */
        time4 = floor (time5 * stiminc / MSEC);
/* fprintf (stderr,"time4 %g otime4 %g\n",time4,otime4); /* */
        if (time4 > otime4) {
                otime4 = time4;
                //runrec(1);              /* run receptors */
        }
    }
    runcomp();                                  /* run compartments */
    if (ptime >= ploti)
     {
      ptime = 1e-8; 
      onplot();				/* run user-defined plot */
      for (i=0; i<numplots; i++) {      /* run user-defined var pen funcs */
		Symbol *vpen;
		int pen;

         if (vpen=(Symbol*)plotnod[i].vpen) {
	    pen = int(callfunc (vpen,3,i,xtime,getrecval(i),0.0));
	    if (pen != plotnod[i].ppen) plotpen (pen,i);
	 }
      }
      for (i=0; i<numplots; i++)                /* display plots */
           if (plotnod[i].pmod != GREC)
		mplot (getrecval(i),xtime,numplots,i,1);
    }

    if (iflag)
     {
       iflag = F;
       break;                           /* stop on interrupt */
     }
  }             /* for (xtime;;) */

  if (stoptime + 2e-10 >= endexp) {
   ptime = 1e-8;
   onplot();					/* run user-defined plot */
   for (i=0; i<numplots; i++)           /* display plots one last time */
      if (plotnod[i].pmod != GREC)
          mplot (getrecval(i),xtime,numplots,i,1);
 }
 runstim(xtime,nsteps,0);		/* run stimuli one more time */
 oldtime = xtime;			/* save time for next runmod() */
 if (runfl && (xtime+ROUNDUP >= endexp)) {
     runyet=0;				/* reset run flag for init */
     runstim(1e20,1,1);      		/* run stimuli one last time */
 }
 else runyet=1;
 timeptr->val = xtime;                  /* save correct time for user */
 timeptr->vtype = NUMBER; 

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"runmod end\n");
#endif
}


/*------------------------------------*/

double record (int cnod1, int cnod2, int cnod3, int cnod4, int pmod, int pval)

{
  node *npnt;
  recep *rpnt;
  double rval, recsynap(int snum, int nf), recchan(int cnum, int nf, int pval);
  comp *cpnt;

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"record node %d %d %d %d mode %d %d",
					cnod1,cnod2,cnod3,cnod4,pmod,pval);
#endif

   rval = 0.0;
   switch (pmod) {

   case VREC: if (!(npnt=findnode(cnod1,cnod2,cnod3,cnod4,
					"record V"))) return(0.0);
	      if (npnt->comptr) rval = npnt->comptr->v;
              break;

   case IREC: if (!(npnt=findnode(cnod1,cnod2,cnod3,cnod4,
					"record I"))) return(0.0);
              if ((cpnt=npnt->comptr)) {
     	         if (cpnt->miscfl & VEXT)             /* comp is v-clamped */
                     rval = -cpnt->extvi;
	      else if (cpnt->miscfl & IEXT)           /* comp is i-clamped */
                     rval = cpnt->exti;
	      }
              break;

   case LREC: if (findrecep (cnod1, cnod2, cnod3, cnod4, &rpnt, "record L"))
                    rval = rpnt->iflux / HMICSEC;
              break;

   case CREC_GLU:  case CREC_AMPA: case CREC_NMDA: case CREC_CNQX:
   case CREC_GABA: case CREC_BIC:  case CREC_PTX:  case CREC_GLY:
   case CREC_STRY:  case CREC_CAMP: case CREC_CGMP: 

         if (!(npnt=findnode(cnod1,cnod2,cnod3,cnod4,
					"record nt"))) return(0.0);
	      if (npnt->comptr) {
		  int typ;

		typ = pmod - CREC_GLU + GLU;
		rval = getnt (npnt->comptr,typ);
	      }
              break;
 
   case CREC_CACONC:
   case CREC_CAV: case CREC_CAIT: case CREC_CAIP: 
   case CREC_CAIE: case CREC_CAIPE: case CREC_CABUF:

         if (!(npnt=findnode(cnod1,cnod2,cnod3,cnod4,
					"record nt"))) return(0.0);
	      if (npnt->comptr) {
		    cacomp *capnt;

		if (capnt=npnt->comptr->capnt) {
		   switch (pmod) {

		case CREC_CABUF: if (capnt->cab) {
		    		  if (pval==0) rval = capnt->cao;
				  else if (pval>(capnt->cashell-1))
				     rval = capnt->cab[capnt->cashell-1];
				  else rval = capnt->cab[pval-1]; 
				 }
				break;
		case CREC_CACONC: 
		    		if (pval==0) rval = capnt->cao;
				else if (pval>(capnt->cashell-1))
				     rval = capnt->cais[capnt->cashell-1];
				else rval = capnt->cais[pval-1]; break;
		    case CREC_CAV:
				rval = capnt->vrev; break;
		    case CREC_CAIT: 
				rval = -capnt->ica;   break;
		    case CREC_CAIPE: 
				rval = -(capnt->ipump+capnt->iexch); break;
		    case CREC_CAIE: 
				rval = -capnt->iexch; break;
		    case CREC_CAIP: 
				rval = -capnt->ipump; break;
		   }
		}
		else {
		  fprintf (stderr,"Node %s does not have a Ca compartment.\n",
			prnode(npnt->nodenm1,npnt->nodenm2,
				npnt->nodenm3,npnt->nodenm4));
		}
	      } 
	      break; 
   case NRECA0: case NRECA1: case NRECA2: case NRECA3: case NRECA4: 
   case NRECA9: 
   case NRECH0: case NRECH1: case NRECH2: case NRECH3: case NRECH4:
   case NRECB0: case NRECB1: case NRECB2: case NRECB3: case NRECB4:
   case NRECC0: case NRECC1: case NRECC2: case NRECC3: case NRECC4:
	  rval = recsynap(cnod1,pmod);
          break;

   case NRECG: case NRECGI: case NRECGV: case NRECGC: case NRECGM: case NRECGH:
   case NRECG0: case NRECG1: case NRECG2: case NRECG3:
	 rval = recchan(cnod1,pmod,pval);
         break;

   }    /* switch */

#ifdef DEBUG
  if (debug & 1) fprintf (stderr," rval %g ...end\n",rval);
#endif
 
 return (rval);
}

/*------------------------------------*/

double recsynap(int snum, int nf)
           
/* Record from a synapse.
    Partly copied from modsynap() in ncmak.c.
*/

{
    elem *elpnt;
    synap *spnt;

    if (!(elpnt=findelem(snum))) {
       fprintf (stderr,"recsyn: can't find element %d\n",snum);
       execerror ("Missing element: "," stopping... ");
       return (0.0);
    }

    spnt = (synap *)elpnt->lptr;	/* get pointer to synapse from elem */
    if (spnt==NULL) {
       fprintf (stderr,"recsyn: can't find synapse for element %d\n",snum);
       return 0.0;
    }
    if (spnt->ctype != SYNAPSE) {
       fprintf (stderr,"recsyn: element %d is not a synapse\n",snum);
       return (0.0);  
    }
    if (spnt->curve==DYAD && nf < NRECB0) {  /* if dyad, use other syanpse */
       if (spnt->dyad) {
           spnt = spnt->dyad->sdyad;/* for presyn filter A */
           if (!spnt) return 0.0;
       }
    }
    if (nf >= NRECA0 && nf < NRECA9) {	/* look at filters before nt release */
      nf = nf - NRECA0;			/* look at any of the NUMFILT filts */
      if      (nf<0) nf = 0;
      else if (nf>=NUMFILT) nf = NUMFILT-1;
      if (spnt->filt1) {
/* fprintf (stderr,"f0 %d snum %d nf %d\n",spnt->filt1->lfilt[nf],snum,nf); */

        return (double)spnt->filt1->lfilt[nf];	/* voltage before filter */
      }
      else return 0.0;
   }
   else if (nf == NRECA9) {		/* look at vesicle rate */
	double trvolts,transrate,vsize;
        int filt;

      if (spnt->filt1) {			/* first check for filter 1 */
            filt = spnt->filt1->nfilt-1;	/* number of filters */
            trvolts = (double)spnt->filt1->lfilt[filt];  /* last filter */
      }
      else trvolts = spnt->comp1->v - spnt->thresh;
      if (spnt->filt1h) {
             filt = spnt->filt1h->nfilt-1;
             trvolts += spnt->filt1h->lfilt[filt] * spnt->filt1h->tfall;
      }
      transrate = getcurv(spnt,trvolts);
      if (transrate < TRMIN) transrate = TRMIN;
      else if (transrate > TRMAX) transrate = TRMAX;

      if ((vsize=spnt->vsize)==0) {
	  if ((vsize=dvsz)==0) vsize = 10;
      }
      transrate /= (vsize*stiminc);
      return transrate;
   }
   else if (nf >= NRECH0 && nf <= NRECH4) {      /* high pass filters */
     nf = nf - NRECH0;			/* look at any of the NUMFILT filters */
     if      (nf<0) nf = 0;
      else if (nf>=NUMFILT) nf = NUMFILT-1;
     if (spnt->filt1h) {
/* fprintf (stderr,"f0h %d snum %d nf %d\n",spnt->filt1h->lfilt[nf],snum,nf); */

       return (double)spnt->filt1h->lfilt[nf];	/* voltage before filter */
     }
     else return 0.0;
   }
   else if (nf < NRECC0) {		/* look at filters after nt release */
      nf = nf - NRECB0;			/* look at any of the NUMFILT filters */
      if      (nf<0) nf = 0;
      else if (nf>=NUMFILT) nf = NUMFILT-1;
      if (spnt->filt2) {
/* fprintf (stderr,"f0 %d snum %d nf %d\n",spnt->filt2->lfilt[nf],snum,nf); */
       return (double)spnt->filt2->lfilt[nf];	/* neurotrans from filter */
      }
      else return 0.0;
   }
   else {	/* nf >= NRECC0 */	/* look at filters after nt sat. */
      nf = nf - NRECC0;			/* look at any of the NUMFILT filters */
      if      (nf<0) nf = 0;
      else if (nf>=NUMFILT) nf = NUMFILT-1;
      if (spnt->filt3) {
/* fprintf (stderr,"f0 %d snum %d nf %d\n",spnt->filt3->lfilt[nf],snum,nf); */
        return (double)spnt->filt3->lfilt[nf];	/* neurotrans from filter */
      }
      else {		/* return normalized conductance if nfilt3 == 0 */
           double c;
        if (nf==0) {
	  if (spnt->resp)           return (spnt->conduct);
	  else if (c=spnt->maxcond) return (spnt->conduct/c);
	  else return  spnt->conduct;
        }
        else return 0.0;
      }
   }
}

/*------------------------------------*/

double recchan(int cnum, int nf, int pval)

			/* cnum = element number of channel */

/* Record from a channel.
    Partly copied from recsynap() above.
*/

{
    elem *elpnt;
    chan *chpnt;
    ntchan *ntchpnt;
    attrib *apnt;
    cacomp *capnt;
    int ncond,nstate;
    double cval;
    synap *spnt;
    recep *rpnt;

 if (!(elpnt=findelem(cnum))) {
      fprintf (stderr,"recchan: can't find element %d\n",cnum);
      execerror ("Missing element: "," stopping... ");
      return 0.0;
 }

 switch (elpnt->ctype) {

  case SYNAPSE:			/* record post-synaptic conductance */

   spnt = (synap *)elpnt->lptr;	/* get pointer to synapse from elem */
   if (spnt==NULL) {
       fprintf (stderr,"recchan: can't find synapse for element %d\n",cnum);
       return 0.0;
   }
  switch (spnt->ctype) {	/* check that it is really a synapse */

   default:
           fprintf (stderr,"recsyn: element %d is not a synapse\n",cnum);
           return (0.0);  
	   break;

   case SYNAPSE:
     if (nf==NRECG0 || (nf==NRECG && pval==0)) { /* total channel conductance */
	if (spnt->resp) cval = spnt->resp->conduct;
	else            cval = spnt->conduct;
       break;
     }

   if ((ntchpnt=(ntchan *)spnt->resp)==NULL) { /* get pnt to chan from elem */
         fprintf (stderr,"recchan: can't find channel for element %d\n",cnum);
         return 0.0;
   }
   else {
   switch (ntchpnt->ctype) {	
     case AMPA:		/* Markov post-synaptic chans */
     case CGMP:	
     case GABA:
     case SYN2:
     case NMDA:
     default:
//      ncond = nf - NRECG0;	/* conductance number */
//      if (ncond > NGREC) ncond = NGREC;
//      else if (ncond < 1) ncond = 1;

      if (pval < 0) pval = 0;

      if ((ntchpnt=(ntchan *)spnt->resp)==NULL) { /* get pnt to chan from elem */
         fprintf (stderr,"recchan: can't find channel for element %d\n",cnum);
         return 0.0;
      }
      switch (nf) {
       double v;

	default: cval = 0; break;
	case NRECGV: cval = ntchpnt->vrev; break;
	case NRECG: 
	  if (pval==0) cval = ntchpnt->conduct;
	  else {
  	    nstate = ntchpnt->numstate;
	    if (pval > nstate) pval = nstate;
	    if (ntchpnt->nchan <= 0)
	       cval = ntchpnt->conc[pval-1].cest;
	    else
	       cval = ((double)ntchpnt->conc[pval-1].nchan) / ntchpnt->nchan;
	  }
	  break;

	  case NRECGI:			/* total ionic current */
	      v = ntchpnt->vrev - ntchpnt->comp1->v;
	      cval = -ntchpnt->conduct * v;
	     break;
	  case NRECGC: 		/* calcium fraction of total conductance */
	      cval = ntchpnt->conduct * ntchpnt->caperm; 
	     break; 

       }  /* switch (ncond) */
       break;

      } /* switch (ntchpnt->ctype) */
    }
    break;

   } /* switch (spnt->ctype) */
    break;

 case GJ: 				/* record gj conductance */
   if (nf==NRECG0) {			/* total channel conductance */
     spnt = (synap *)elpnt->lptr;	/* get pointer to synapse from elem */
     cval = spnt->conduct;
   }
    break;

 case ROD:
 case CONE:
    rpnt = (recep *)elpnt->lptr;	/* get pointer to photorec from elem */
    switch (nf) {
      default: cval = 0; break;
      case NRECG:
      if (pval > 10) pval = 10;
      switch (pval) {

        case 0: cval = rpnt->cond;  break;
        case 1: cval = rpnt->rhod;  break;
        case 2: cval = rpnt->rhod2; break;
        case 3: cval = rpnt->rstar; break;
        case 4: cval = rpnt->gpr;   break;
        case 5: cval = rpnt->pde;   break;
        case 6: cval = rpnt->gcyc;  break;
        case 7: cval = rpnt->cycg;  break;
        case 8: cval = rpnt->ca;    break;
        case 9: cval = rpnt->cax;   break;
        case 10:cval = rpnt->rhodk; break;
      }
    break;
   }
   break;

 default:
 case CHAN:			/* record channel conductances */
 	
  if ((apnt=elpnt->attpnt)==NULL) {
  fprintf (stderr,"recchan: can't find channel attrib for element %d\n",cnum);
       return 0.0;
  }

  if ((chpnt=(chan *)apnt->lptr)==NULL) { /* get pnt to channel from elem */
       fprintf (stderr,"recchan: can't find channel for element %d\n",cnum);
       return 0.0;
  }

/*  fprintf(stderr,"chan type %d stype %d nf %d\n",
		chpnt->ctype,chpnt->stype,nf); /* */

  switch (chpnt->ctype) {	/* stop short on wrong element */
     case AMPA:			/* synaptic Markov chans */
     case GABA:
     case SYN2:
     case NMDA:

     case CGMP:
     case NA:
     case K:
     case KCa:
     case CA: break;
     default:
       fprintf (stderr,"recchan: element %d is not a channel (%d)\n",
					cnum,chpnt->ctype);
       return (0.0);  
  }

  if (nf==NRECG0) {		/* total channel conductance */
     cval = chpnt->conduct;
  }
  else {    
      double v;

//    ncond = nf - NRECG0;	/* conductance number */
//    if (ncond > NGREC) ncond = NGREC;
//    else if (ncond < 1) ncond = 1;

    switch (nf) {
        double nernstv(double val);

		/* params that are the same for all chans */

    case NRECGI: 
		 switch (chpnt->ctype) {
		   case CA: v = chpnt->gvrev; break;
		   default: v = chpnt->vrev; break;
		 }
		 v -= chpnt->comp1->v;
		 cval = chpnt->conduct * v;
		 break;
    case NRECGV: switch (chpnt->ctype) { 
			chantype *chtyp;

		  case NMDA:
		  case CGMP:
		  case CA:
			if (chpnt->setvrev) cval = chpnt->vrev;
			else { 
			  chtyp = getchantype(chpnt->ctype,chpnt->stype);	
			  if (capnt=chpnt->comp1->capnt) 
			     cval = ghkv(chtyp,capnt->cao,capnt->cais[0]);

			  else cval = ghkv(chtyp,dcao,dcai);
			}
			break; 
		  default: cval = chpnt->vrev; break;
		}
		break;
    case NRECGC: cval = chpnt->conduct * chpnt->caperm; break;

    case NRECG:  if (pval==0) {
		 cval = chpnt->conduct;
		 break; 
		 }
		/* else if (pval > 0), continue on to next case: */

		/* params that are different for channel subtypes */
    default:
    switch (chpnt->ctype) {
     case NA:
	  switch (chpnt->stype) {
	    case 0:			/* m3h type Na channel */
	 	switch (nf) {
		    case NRECG1:
		    case NRECGM: cval = ((hhchan *)chpnt)->m; break;
		    case NRECG2:
		    case NRECGH: cval = ((hhchan *)chpnt)->h; break;
    		    case NRECG3: v = chpnt->vrev - chpnt->comp1->v;
		                  cval = -chpnt->conduct * v; break;
		    default:     cval = 0.0; break;
		}
		break;
	    case 1:
	    case 2:
	    case 3:
	    default:
	      switch (nf) {
	        default: cval = 0; break;
		case NRECG:
		nstate = chpnt->numstate;
		if (pval > nstate) pval = nstate;
		if (chpnt->nchan <= 0)
		  cval = chpnt->conc[pval-1].cest;
		else
		  cval = ((double)chpnt->conc[pval-1].nchan) / chpnt->nchan;
		break;
		}
	  }
 	  break;
     case K:
	  switch (chpnt->stype) {
	    case 0:			/* n4 type K channel */
	 	switch (nf) {
		    case NRECG1:
		    case NRECGM: cval = ((hhchan *)chpnt)->m; break; /* n */
    		    case NRECG2: v = chpnt->vrev - chpnt->comp1->v;
		                  cval = -chpnt->conduct * v; break;
		    default:     cval = 0.0; break;
		}
		break;
	    case 1:
	    case 3:			/* KA, markov */
	    case 4:			/* Ih */
	    default:
	      switch (nf) {
	        default: cval = 0; break;
		case NRECG:
		nstate = chpnt->numstate;
		if (pval > nstate) pval = nstate;
		if (chpnt->nchan <= 0)
		  cval = chpnt->conc[pval-1].cest;
		else
		  cval = ((double)chpnt->conc[pval-1].nchan) / chpnt->nchan;
		break;
	      }
	      break;
	    case 2:			/* KA, HH */
	 	switch (nf) {
		    case NRECG1:
		    case NRECGM: cval = ((hhchan *)chpnt)->m; break; /* n */
		    case NRECG2:
		    case NRECGH: cval = ((hhchan *)chpnt)->h; break; /* h */
    		    case NRECG3: v = chpnt->vrev - chpnt->comp1->v;
		                  cval = -chpnt->conduct * v; break;
		    default:    cval = 0.0; break;
		}
		break;
	  }
 	  break;
     case KCa:
	  switch (chpnt->stype) {
	    case 0:			/*  sKCa channel */
	    case 2:			/*  bKCa channel */
	 	switch (nf) {
		    case NRECG1:
		    case NRECGM: cval = ((hhchan *)chpnt)->m; break; /* n */
    		    case NRECG2: v = chpnt->vrev - chpnt->comp1->v;
		                  cval = -chpnt->conduct * v; break;
		    default:     cval = 0.0; break;
		}
		break;
	    case 1:			/* sKCa */
	    case 3:			/* bKCa */
	    case 4:			/* sKCa */
	    case 5:			/* sKCa */
	    case 6:			/* bKCa */
	    default:
	      switch (nf) {
	        default: cval = 0; break;
		case NRECG:
		nstate = chpnt->numstate;
		if (pval > nstate) pval = nstate;
		if (chpnt->nchan <= 0)
		  cval = chpnt->conc[pval-1].cest;
		else
		  cval = ((double)chpnt->conc[pval-1].nchan) / chpnt->nchan;
		break;
	      }
	      break;
	  }
 	  break;
     case CA:
	capnt = chpnt->comp1->capnt;
	switch (chpnt->stype) {
        case 0:			/* m3 type Ca channel */
        case 1:			/* m3 type Ca channel */
 	switch (nf) {
	    case NRECG1:
	    case NRECGM: cval = ((hhchan *)chpnt)->m; break; /* m */
	    case NRECGH: cval = ((hhchan *)chpnt)->h; break; /* h */
    	    case NRECG2: v = chpnt->vrev - chpnt->comp1->v;
	                  cval = -chpnt->conduct * v; break;
	    default:    cval = 0.0; break;
		}
		break;
	 case 2:
	 default:
	  switch (nf) {
	    default: cval = 0; break;
	    case NRECG:
	    nstate = chpnt->numstate;
	    if (pval > nstate) pval = nstate;
 	    if (chpnt->nchan <= 0)
	       cval = chpnt->conc[pval-1].cest;
	    else
	       cval = ((double)chpnt->conc[pval-1].nchan) / chpnt->nchan;
	    break;
	    }
	  break;
	}
        break;

     default:			/* any other channel type */
     case CGMP:
	  switch (chpnt->stype) {
	  case 1:
	  default:
	    switch (nf) {
	      default: cval = 0; break;
	      case NRECG:
	      nstate = chpnt->numstate;
	      if (pval > nstate) pval = nstate;
	      if (chpnt->nchan <= 0)
		  cval = chpnt->conc[pval-1].cest;
	      else
		  cval = ((double)chpnt->conc[pval-1].nchan) / chpnt->nchan;
              break;
	     }
 	  break;
    }  /* switch (chpnt->stype) */
    break;
    
    }  /* switch (chpnt->ctype) */
   }  /* switch (nf)    */
  } /* else nf > NRECG0 */

  break; 	/* case CHAN */

 }  /* switch (elpnt->ctype) */

 return cval;
}

/*------------------------------------*/

double reccable(int elemnum, double fracdist)

/* record from a cable with element number "elemnum",
   at a fractional distance "fracdist" from the first
   end (the end that contacts node 1, the node given first).
*/

{
    int i,n,cnum;
    double r, v, vold;
    node *np1;
    elem *epnt;
    conn *cpnt,*cpnt1,*cpnt2,*tcpnt;
    comp *pnt,*pnt1,*pnt2;
    conlst *lpnt;


 if (!(epnt=findelem(elemnum))) {
	fprintf (stderr,"reccable: can't find element %d\n",elemnum);
	execerror ("Missing element: "," stopping... ");
	return 0.0;  
 }

 np1 = epnt->nodp1;
 if (!np1) {
         fprintf (stderr,"reccable: can't find node for element %d\n",elemnum);
         return 0.0;  
 }
 cpnt = (conn *)epnt->lptr;
 if (!cpnt) {
         fprintf (stderr,"reccable: can't find low-lev conn for element %d\n",
							elemnum);
  	 return 0.0;
 } 
 		/* find first node's compartment */

 pnt = np1->comptr;
 if (!pnt) {
      fprintf (stderr,"reccable: can't find comp for node %d\n",np1->nodenm1);
      return 0.0;
 } 
 		/* Find connection from first comp to second comp. */ 
		/* Must be same as connection saved in cable. */

 cpnt1 = (conn*)NULL;
 for (lpnt=pnt->clst; lpnt && (cpnt1 != cpnt); lpnt=lpnt->next) {
    if (!(cpnt1=lpnt->conpnt)) {
	fprintf (stderr,"reccable: missing conn pt\n");
	return 0.0;
    }
 }
 tcpnt = cpnt1;		/* remember the conn */

		/*  (Code below taken from dcomp() in "prcomp.c") */

		/* Make sure conn points back to orig comp. */

 if 	 ((pnt1=cpnt1->comp1)==pnt) pnt1 = cpnt1->comp2;
 else if ((pnt1=cpnt1->comp2)==pnt) pnt1 = cpnt1->comp1;
 else return 0.0;

 if (cpnt1->ctype!=AXIALRES) return 0.0;
 
 for (n=0; pnt1->nodlst==NULL; n++) {	       /* traverse cable, count comps */

    lpnt = pnt1->clst;
    if (!lpnt) break;
    if (!(cpnt2=lpnt->conpnt)) break;		 /* find conn */
    if (cpnt1==cpnt2) {
	lpnt  = lpnt->next;
	if (lpnt) cpnt2 = lpnt->conpnt; 	 /* wrong conn? */
    }
    if (!cpnt2) break;
    while (cpnt2->ctype!=AXIALRES) {
      lpnt  = lpnt->next;
      if (!lpnt) break;
      cpnt2 = lpnt->conpnt;
      if (!cpnt2) break;
    }

    if ((pnt2=cpnt2->comp1)==pnt1) pnt2 = cpnt2->comp2;
    pnt1=pnt2;
    cpnt1=cpnt2;

   }   /* for (n=0;;) */

   n++;					/* final comp at second node */
   cnum = (int)(fracdist * n);		/* integer number of requested comp */
  					/* end comps (0,n) are 1/2 size */
 
			/* Do it again; this time stop at comp "cnum" */
   cpnt1 = tcpnt;
   if 	   ((pnt1=cpnt1->comp1)==pnt) pnt1 = cpnt1->comp2;
   else if ((pnt1=cpnt1->comp2)==pnt) pnt1 = cpnt1->comp1;
   else return 0.0;

   if (cpnt1->ctype!=AXIALRES) return 0.0;

   vold = pnt->v;
   v = pnt1->v;
   for (i=1; i<=cnum; i++) {	       /* traverse cable, count comps */

      lpnt = pnt1->clst;
      if (!(cpnt2=lpnt->conpnt)) break;		 /* find conn */
      if (cpnt1==cpnt2) {
	  lpnt  = lpnt->next;
	  if (lpnt) cpnt2 = lpnt->conpnt; 	 /* wrong conn? */
      }
      if (!cpnt2) break;
      while (cpnt2->ctype!=AXIALRES) {
        lpnt  = lpnt->next;
        if (!lpnt) break;
        cpnt2 = lpnt->conpnt;
        if (!cpnt2) break;
      }

      if ((pnt2=cpnt2->comp1)==pnt1) pnt2 = cpnt2->comp2;
      vold = v;
      v = pnt2->v;
      pnt1=pnt2;
      cpnt1=cpnt2;

     }   /* for (i=0;;) */

   r = fracdist * n - cnum;		/* interpolate across axial resistors */
					/* between adjacent compartments */

/*  fprintf (stderr,"n %d fracdist %g cnum %d v %g vold %g r %g\n",
			n,fracdist,cnum,v,vold,r);	/* */
    
   return (vold * (1.0-r) + v * r);
}

/*------------------------------------*/

double getrecval (int plotnum)
{
  int cnod1,cnod2,cnod3,cnod4,pmod,pval;
  Symbol *spnt;
  int narg;
  short *vtyp;
  double *val;
  double rval, *garrayval(Symbol *s);
  double recsynap(int snum, int nf),reccable(int elemnum, double fracdist);

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"getrecval plotnum %d\n",plotnum);
#endif

   cnod1 = smaxmin(plotnod[plotnum].cnod1);
   cnod2 = smaxmin(plotnod[plotnum].cnod2);
   cnod3 = smaxmin(plotnod[plotnum].cnod3);
   cnod4 = smaxmin(plotnod[plotnum].cnod4);

   pmod  = plotnod[plotnum].pmod;
   pval  = plotnod[plotnum].pval;

   switch (pmod) {

   case VREC:
   case IREC:
   case LREC: 
   
   case CREC_GLU:  case CREC_AMPA: case CREC_NMDA: case CREC_CNQX:
   case CREC_GABA: case CREC_BIC:  case CREC_PTX:  case CREC_GLY:
   case CREC_STRY:  case CREC_CAMP: case CREC_CGMP: 

   case CREC_CACONC:
   case CREC_CAV: case CREC_CAIT: case CREC_CAIP: case CREC_CAIE: 
   case CREC_CAIPE: case CREC_CABUF:

             rval = record (cnod1,cnod2,cnod3,cnod4,pmod,pval);
             break;
   case GREC:
	     break;
   case SREC: 
        spnt = (Symbol *)plotnod[plotnum].spnt;  /* code is from getvar() */
        narg = (long int)spnt;
        if (narg > 0 && narg < 20) {    /* if var is arg to a function */
                getarg(narg,&val,&vtyp);   
		rval = *val;
        }
        else {
          switch (spnt->type) {
          case UNDEF:
          case CONST:
          case VAR:    rval = spnt->val;
                      break;
          case ARRAY:  
                      rval = *garrayval(spnt); 
                      break;
          default: execerror("attempt to eval non-var", spnt->name);
                      break;
          } /* switch */
        } /* else */
        break;
			/* record from synapse */

case NRECA0: case NRECA1: case NRECA2: case NRECA3: case NRECA4:
case NRECA9:
case NRECH0: case NRECH1: case NRECH2: case NRECH3: case NRECH4:
case NRECB0: case NRECB1: case NRECB2: case NRECB3: case NRECB4:
case NRECC0: case NRECC1: case NRECC2: case NRECC3: case NRECC4:

			/* record conductance from channel */

case NRECG0: case NRECG1: case NRECG2: case NRECG3:
case NRECG: case NRECGI: case NRECGV: case NRECGC: case NRECGM: case NRECGH: 

			/*  cnod1 = element number, not node number */ 
		rval = record(cnod1,cnod2,cnod3,cnod4,pmod,pval);
        	break;
case CABLE:
     		rval = reccable(cnod1,((double) cnod2)/CABLFRAC);
		break;

   }    /* switch */
 
#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"getrecvalend\n");
#endif

 return (rval);
}

/*------------------------------------*/

double getcurv(synap *spnt, double val)

/* Given a synapse (pointer) and a voltage (volts), return
instantaneous value for amount of neurotransmitter released per
time step (100 usec), in numbers of half-saturating vesicles
(i.e. vsize=1). */

{
   double rval=0;

  switch (spnt->curve) {

   case LINEAR:	if (val < 0) val = 0;		/* linear, sharp shoulder */
		rval = val * 10.0 * spnt->ngain;/* val=0.1V -> rval = 1 */
		break;

   case EXPON:	rval =  .025 * exp(val*VTOMV/spnt->ngain); /* exponential */
		if (rval < .00001) rval = 0.0;	/* val is in volts */ 
		break;
  }
/*  fprintf (stderr,"curv %d, ngain %g val %g, rval %g cgain %g\n",
			spnt->curve,spnt->ngain,val,rval,spnt->cgain); /* */
 return (rval);
}

/*------------------------------------*/

void runsyn(int stime)
  			/* number of 100 microsec steps */

/* Run all the synapses, depending on their
   type and history.  Normally, call once every 100 microsec.
   For static runs, call with stime=0 and only
   the amplitude transfer function is produced, not the
   time functions.
*/

{
   int i,nsites;
   double trvolts,fltvolts,vesrel,transrate,p;
   double vsize,refr,meanves,pmean,nvtime;       
   double chanligand,bnldev(double pp, int n);
   double synfilt(lpfilt *lpnt, double transrate);
   synap *spnt;
   dyadlst *lpnt;
   gj *gjpnt;
   double vj, cyc, normcond;
   double alpha, beta, n, rpar, gnv, sd;
   comp *cpnt;

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"runsynap");
#endif

 for (spnt=synpnt; spnt; spnt= (synap *)spnt->next)
   {
     switch (spnt->ctype)
      {
       case SYNAPSE:
	 if (spnt->curve==DYAD) break;		/* stop if 2nd in dyad */
	 if (stime) {
	    trvolts = (spnt->comp1->v - spnt->thresh);  /* calib in volts */
            for (i=0; i<stime; i++) {                   /* one or more steps */
	       if (spnt->filt1)
	         fltvolts = synfilt(spnt->filt1,trvolts); /* first lowp filt */
	       else fltvolts = trvolts;
	       if (spnt->filt1h)		/* highpass filt */
	         fltvolts += synfilt(spnt->filt1h,trvolts); 
	    }
	 }
	 else {
            trvolts = (spnt->comp1->vest - spnt->thresh); /* calib in volts */
	    fltvolts = trvolts;				/* if static */
	 }
	 transrate = getcurv(spnt,fltvolts);         /* static tranfer func */
         if (transrate < TRMIN) transrate = TRMIN;
         else if (transrate > TRMAX) transrate = TRMAX;
	 if (stime) {
           for (i=0; i<stime; i++) {                   /* one or more steps */
 	     if (nsites=spnt->vsites && spnt->vsize) {
	        vsize=spnt->vsize * HMICSEC/stiminc;   /* ves rate*timeinc */

				/* for binomial, stdev = sqrt(np(1-p)) */

	        meanves = transrate / vsize; 
		sd = meanves * spnt->cov; 
		refr = spnt->vrefr;
		if (refr < 0.0) refr = 0.0;
		if ((refr > 1) && (meanves > 1)) {
		    meanves = 1; 
		}
	        if (meanves > 1) {
	          if (spnt->cov > 0.3) {
	            p = 1e-2;
		  }
		  else {
		    p = 1.0 - (sd * sd / meanves);
		  }
		  nsites = (int) (meanves / p);
		  if (nsites < 1) nsites = 1;
		  p = meanves / nsites;
		  if (p < 0) p = 0;
		  else if (p > 1.0) p = 1.0;
		  if (spnt->vstate) rsetstate (spnt->vstate);
	          vesrel = bnldev(p,nsites);    
		}
		else {		/* Less than 1 ves/time step, */
				/*  use refractory period: */
		   int gorder;

		/* Mark van Rossum's way for calculationg refr period: */

/*		   refr = 1.0 / meanves - sd / (meanves*sqrt(meanves)); */

		   gorder = int(1.0 / (spnt->cov*spnt->cov)+0.5);
		   if (meanves<MINVES) {	/* If very low rate, */
		      meanves = MINVES;		/*  just reset */
		      pmean = 1.0 / meanves;
		      if (spnt->vtime > 0) spnt->vtime = 0;
		      if (spnt->vstate) rsetstate (spnt->vstate);
		      spnt->vtint = gamdev(gorder)/gorder;
		      spnt->vflag = 0;
		      vesrel = 0; 
		   }
		   else {	/* meanves is non-zero */
			double x;
		     x = 1.0 / meanves;
		     if (spnt->vflag) x -= refr;
		     if (x >= 0) { 	     /* asymptote to x for pos pmean */
			  if (x > 100) pmean = x;
		          else         pmean = exp(-x/2) + x;
		     }
		     else {		     /* asymptote to 0 for neg pmean*/ 
			if (x < -100) pmean=0; 
		        else          pmean = exp(x/2);
		     }
		     if (pmean > MAXVINT) pmean = MAXVINT; 	

/* fprintf (stderr,"vtime %g vtint %g pmean %g meanves %g gorder %d refr %g\n",
  spnt->vtime, spnt->vtint, pmean, meanves,gorder, refr);  /* */

				/* if rate increases, reset interval */

							/* vesicle event */
		   if (spnt->vtime >= spnt->vtint) {
		         if (spnt->vstate) rsetstate (spnt->vstate);
		         spnt->vtint = gamdev(gorder)/gorder;
		         spnt->vtime = -refr / pmean;
		         spnt->vflag = 1;
		         vesrel = 1;
		   }
		   else {		/* no event, just count interval */
		      vesrel = 0;
		    }

		  } /* else meanves is nonzero */

		  nvtime = spnt->vtime * pmean;
		  nvtime++;
      		  if (nvtime > MAXVINT) {
		      spnt->vtime = -refr / pmean;	
		  }
		  spnt->vtime = nvtime / pmean;

/* fprintf (stderr, "vtime %g vtint %g meanves %g pmean %g vesrel %g\n",
		spnt->vtime,spnt->vtint,meanves,pmean,vesrel); /* */

		}  /* less than 1 ves/step */

/* fprintf (stderr, "vtime %g vtint %g meanves %g vesrel %g\n",
		spnt->vtime,spnt->vtint,meanves,vesrel); /* */

		transrate = vesrel * vsize; 

			/* The orig Poisson way: */

//	/*	   if (spnt->vstate) rsetstate (spnt->vstate);
//		   vesrel = poisdev (transrate/vsize); 
//		   transrate = vesrel * vsize; 
//	*/
//
//			/* the old binom way: */
//
//  /*  #define TMAXP 5.0     			     /* max prob at 10.0 */ 
//    /*		   transrate /= TMAXP;
//		   if ((p=transrate/(vsize*nsites))<=1) {
//		     if (vsize<1) {		   
//		       p = transrate/nsites;	    
//		       nsites = (int)(nsites / vsize);
//		     }
//		     nsites = (int)(nsites * TMAXP);
//		     if (spnt->vstate) rsetstate (spnt->vstate);
//	             vesrel = bnldev(p,nsites);    
//		     transrate = vesrel * vsize; 
//		   }
//		   else {
//		     transrate *= TMAXP;
//		   }
//	*/

/* fprintf(stderr,"vesrel %g transrate %g p %g size %g nsites %d\n",
			vesrel,transrate,p,vsize,nsites); 	/* */
	 	}
/* fprintf(stderr,"transrate %g\n", transrate); 	/* */
	       if (spnt->filt2)
		 chanligand = synfilt(spnt->filt2,transrate); /* 2nd lowp filt*/
	       else chanligand = transrate;
               if (chanligand < TRMIN) chanligand = TRMIN;
               else if (chanligand > TRMAX) chanligand = TRMAX;
            }
	    restorstate();    /* restore random state from bnldev and gamdev */
	 }  /* if (stime) */
	 else chanligand = transrate;

         rchan(spnt,stime,chanligand);		/* run the channels */
         if (spnt->dyad) {			/* run dyad's 2nd filt, chans */
	   for (lpnt=spnt->dyad; lpnt; lpnt=lpnt->next) {
             if (lpnt->sdyad) {
               if (lpnt->sdyad->filt2) 
		 chanligand = synfilt(lpnt->sdyad->filt2,transrate); 
               else chanligand = transrate;
               if (chanligand < TRMIN) chanligand = TRMIN;
               else if (chanligand > TRMAX) chanligand = TRMAX;
               rchan(lpnt->sdyad,stime,chanligand); /* run the dyad's channels*/
	     }
           }
 	 }
         break;   /* case SYNAP */

       case GJ:                                 /* GJ modulated by voltage */
	gjpnt = (gj *)spnt;
	gnv = gjpnt->gnv;
	if (stime) {
         for (i=0; i<stime; i++) {		/* one or more steps */

	cyc = getmesng(gjpnt,gjpnt->modtyp);

	if (gjpnt->sign) cyc = 1.0 - cyc;

			/* For details, see Harris, Spray, Bennett, */
			/* J.  Gen Physiol. 77:95-117, 1981.        */

#define GJLAMBDA 0.0013
#define Aa 0.077
#define Ab 0.14

          vj = gjpnt->comp1->v - gjpnt->comp2->v;
	  alpha = gjalpha(vj, gjpnt);
	  beta =  gjbeta (vj, gjpnt);
	  n = gjpnt->n;
	  n += -beta * n + alpha * (1.0 - n);
	  if (n < 0.0) n = 0.0;
	  if (n > 1.0) n = 1.0;
	  gjpnt->n = n;
          gjpnt->conduct = gjpnt->maxcond * ( gnv + n * (1.0-gnv)) * (1.0-cyc);

 	  }   /* for (i;;) */ 
	 }  /* if (stime;;) */

	 else { 	/* stime==0,  static gj, update compartment implf */

	   cyc = getmesng(gjpnt,gjpnt->modtyp);
	   if (gjpnt->sign) cyc = 1.0 - cyc;

           vj = ncabs(gjpnt->comp1->v - gjpnt->comp2->v);
	   alpha = gjalpha(vj, gjpnt);
	   beta =  gjbeta (vj, gjpnt);
	   n = alpha / (alpha + beta);
	   gjpnt->n = n;
           normcond = (gnv + n * (1.0-gnv)) * (1.0-cyc);

	   cpnt = gjpnt->comp1;		/* check first comp */
	   if (cpnt) {
	     if (cpnt->implf) {	/* if implf has been calculated yet... */
	      rpar = ((1. / cpnt->implf) - 1.) / cpnt->k - gjpnt->conduct;
	      rpar += normcond * gjpnt->maxcond; 
	      cpnt->implf = 1. / (1. + rpar * cpnt->k);  /* implicit factor*/
	    }
	  }

	   cpnt = gjpnt->comp2;		/* check second comp */
	   if (cpnt) {
	     if (cpnt->implf) {	/* if implf has been calculated yet... */
	      rpar = ((1. / cpnt->implf) - 1.) / cpnt->k - gjpnt->conduct;
	      rpar += normcond * gjpnt->maxcond; 
	      cpnt->implf = 1. / (1. + rpar * cpnt->k);  /* implicit factor*/
	    }
	  }

          gjpnt->conduct = gjpnt->maxcond * normcond;
	 } /* else static */
         break;

       case BUF:                                /* delay buffer */
        for (i=0; i<stime; i++) {                   /* one or more steps */
	   if (((dbuf *)spnt)->delay > 0) {	/* if delay */
		int v;
		double vf;
   	        dbuf *dlpnt;

	      dlpnt = (dbuf *)spnt;
	      vf = dlpnt->comp1->v / DELCONV; /* input voltage */
	      if (vf > 0) 
	         v = (int)(vf+0.5); 		/* input voltage */
	      else
	         v = (int)(vf-0.5); 		/* input voltage */
	      if (v>MAXSHORT) v = MAXSHORT;	/* no overflow of short int */
	      else if (v<MINSHORT) v = MINSHORT;
	      if (dlpnt->delbuf) {
	        *dlpnt->delpnt = v;
	        if (++dlpnt->delpnt >= dlpnt->delbuf+ (int)(dlpnt->delay))
		   dlpnt->delpnt = dlpnt->delbuf;  /* incr. pntr, make circ. */
					/* leave pointing to next output val */
	      }  /* if (dlpnt->delbuf) */
	    }  /* if (dlpnt->delay>0) */
	    else {
		((dbuf *)spnt)->comp2->v = ((dbuf *)spnt)->comp1->v;
	    }
	 }  /* if (stime;;) */
         break;  /* case BUF */

      } /* switch */
   }        /* for */
#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"...end\n");
#endif
}

/*------------------------------------*/

double setbind(synap *spnt, double chligand)

/* calc receptor binding and norm. conductance for standard synapse */

{
   int i;
   double bound,k,kk,ll,pp;

 pp = spnt->npow;		/* Hill coefficient */
 k = spnt->nkd;
 if (floor(pp) == pp) {         /*  integer power */
    kk = k;
    ll = chligand;
    for (i=int(pp); i>1; i--) {
       kk *= k;
       ll *= chligand;
    }
 }
 else {                       /* non-integer power */
     kk = pow(k,pp);
     ll = pow(chligand,pp);
 }
 bound = ll / (ll + kk);  	/* fraction bound */

 return bound;
}

/*------------------------------------*/

double setcond(synap *spnt, double bound)

/* calc conductance for standard synapse */

{
   double conduct,cycg,cycgmax;
   comp *cpnt;

 switch (spnt->ntact) {		/* calc conductance */
     case OPEN:
         conduct = bound;
         break;

  /* Here we simulate the action of the cGMP cascade in an
     on-bipolar.  See Shiells and Falk, 1994. */

     case CLOSE:
      if (cpnt=spnt->comp2) {
        if (spnt->mesng1 || spnt->resp) {	/* if special response */
            cycgmax = 1.0;
            cycg = cycgmax - bound * spnt->cgain;
            if (cycg < 0.0) cycg = 0.0;
	    conduct = cycg;	    
        }
	else {  				/* (!spnt->mesng) */
	   if (isnt(cpnt,CGMP)) cycgmax = getnt(cpnt,CGMP);
	   else                 cycgmax = 1.0;
           cycg = cycgmax - bound * spnt->cgain;
           if (cycg < 0.0) cycg = 0.0;
           conduct = cycg / (cycg + spnt->ckd) * (1 + spnt->ckd);
       }
    }
    break;

 } /* switch (spnt->ntact) */

return conduct;
}

/*------------------------------------*/

void rchan(synap *spnt, int stime, double chligand)

/* Run the channel conductance part of a synapse.  */

/*  AMPA, CGMP, GABA, SYN2, and NMDA channels are run as state machines 
    instead of having binding and temporal filters. Their conductances 
    are found and assigned to the synaptic conductance just like
    the standard synapse.  These channels are run during the 
    compartment iteration so are not part of this procedure.
*/

{
   int i,nchan;
   double bound,active,cycg,cdur,rpar;
   double synfilt(lpfilt *lpnt, double transrate);
   double k,kk,ll,pp,cycgmax;
   comp *cpnt;


  if (spnt->resp) {			/* if postsynaptic markov channel */
    switch (spnt->resp->ctype){		/* calc "active" signal sent to chan */

	case AMPA:			/* ligand-binding channels */ 
	case NMDA: 
	case GABA: active = chligand; break;

	case CGMP:			/* "second-messenger" chans */
	case SYN2:
   	  bound = setbind (spnt,chligand);		/*  calc binding */
   	  if (stime && spnt->filt3) bound = synfilt(spnt->filt3,bound);
   	  active = setcond(spnt,bound);     /* calc norm cond, 2nd mesng  */
	  break;
    }
  }
  else {			/* else standard postsynaptic mechanism */

   bound = setbind (spnt,chligand);		/*  calc binding */
   if (stime && spnt->filt3) bound = synfilt(spnt->filt3,bound);
   active = setcond(spnt,bound);     /* calc norm cond, possib w/2nd mesng */

   if (!stime) { /* stime==0,  static syn, update compartment implf */
     cpnt = spnt->comp2;
     if (cpnt) {
       if (cpnt->implf) {	/* if implf has been calculated yet... */
         rpar = ((1. / cpnt->implf) - 1.) / cpnt->k - spnt->conduct;
         rpar += active * spnt->maxcond; 
         cpnt->implf = 1. / (1. + rpar * cpnt->k);  /* implicit factor*/
       }
     }
     else fprintf (stderr,"rchan: panic, no second compartment\n");
   }
  } 

  if (spnt->mesng1){
       double a;

     a = (active - spnt->oldc)*spnt->trconc;
     setnt(spnt->comp2,spnt->mesng1, a);
     if (spnt->mesng2) setnt(spnt->comp2,spnt->mesng2, a);
     spnt->oldc = active;
  }

  if (spnt->resp)
     spnt->conduct = active; /* save norm. conductance for syn2 chan */
   else
     spnt->conduct = active * spnt->maxcond;

/* fprintf (stderr,"chlig %g bound %g nkd %g ntact %d\n",
			chligand,bound,spnt->nkd,spnt->ntact);  */
}

/*------------------------------------*/

void runrec(int dtimestep)

/* Run all the receptors. 
 depending on their type and history.
 Call at multiples of tenths of milliseconds only. */

{
   int i,tspeed;
   recep *rpnt;
   recpar *cpt;
   static double iflux,pflux,rflux,ca2;
   double drand(void);
   void restorstate();

#ifdef DEBUG
  if (debug & 1)  fprintf (stderr,"runrec");
#endif

  for (rpnt = recpnt; rpnt; rpnt=(recep *)rpnt->next) {

 
     switch (rpnt->ctype) {

       case TRANSDUCER:
	  rpnt->iflux = rpnt->flux*HMICSEC; 	/* save iflux for record */
	  break;
       case ROD:
       case CONE:
          if (!(cpt=rpnt->chtyp)) {		/* get pointer to const */
	     break;
	  }
	  else {
	  if (tempcel != cpt->oldtempcel) initrec(); /* recalc if temp changes */

	  tspeed = (int) (rpnt->tspeed * dtimestep);
	  rflux = rpnt->flux * HMICSEC;
          if (rpnt->pnois) {
                if (rpnt->pstate) rsetstate (rpnt->pstate);
		pflux = poisdev(rflux);
		restorstate();
	        iflux = pflux;
	  }
	  else iflux = rflux; 
	  rpnt->iflux = iflux;  		/* save iflux for record */
	  iflux *= rpnt->intenadj;		/* adjust inten for timec */
          for (i=0; i<tspeed; i++) {             /* maybe do several steps */
		double t1, t2, t3;

            rpnt->rhod  += iflux * cpt->lgain;                /* */

	    t1 = rpnt->rhod  * cpt->rgain1;
	    rpnt->rhod  -= t1;
            rpnt->rhod2 += t1;

            t2 = rpnt->rhod2 * cpt->rgain2;
            rpnt->rhod2 -= t2;
            rpnt->rstar += t2;

            rpnt->gpr   +=  rpnt->rstar * cpt->rgain3;

            t3 = rpnt->gpr * cpt->ggain;
	    rpnt->gpr -= t3; 
	    rpnt->pde += t3;

	    if (rpnt->dnois) {
                 if (rpnt->dstate) rsetstate (rpnt->dstate);
		 rpnt->pde  += (drand() - 0.5) * .0008 * rpnt->dnois;
 		 restorstate();  	/* restore default random state */
	    }

	      /* Calcium feedback */

            ca2 = rpnt->cax;
            ca2 = ca2 * ca2;

            rpnt->gcyc  = 1.0 - (ca2 / (ca2 + cpt->kdgcyc));
	    rpnt->rhodk = 1.0 - (ca2 / (ca2 + cpt->kdrhodk));

            rpnt->cycg -= rpnt->pde * rpnt->cycg * cpt->pdegain;
            rpnt->cycg += rpnt->gcyc * cpt->gcygain;
            if (rpnt->cycg < 0) rpnt->cycg = 0;

            rpnt->cond += rpnt->cycg * rpnt->cycg * rpnt->cycg *
                         (cpt->totcond - rpnt->cond) /
                         cpt->totcond * cpt->condgain;
            rpnt->ca   += rpnt->cond * cpt->gca;
            rpnt->cax  += rpnt->ca * cpt->cxgain;

            rpnt->rstar *= cpt->decrstar - rpnt->rhodk * cpt->rkgain;	

            rpnt->pde  = cpt->pdedec * (rpnt->pde- cpt->pdebase) + cpt->pdebase;
            rpnt->ca   -= cpt->capump * rpnt->ca / (rpnt->ca + 1.0);
            rpnt->cond *= cpt->deccond;
            rpnt->cax  *= cpt->deccax;
   
 	  }	/* for */
        }    /* else */

         rpnt->conduct = rpnt->maxcond * rpnt->cond;   /* chan current */

if (debug & 1 && debugz & 16)
 fprintf (stderr,
        "rec num %d tspeed %g rpntinten %g rflux %g pflux %g iflux %g flux %g pnois %d rh %g pd %g gc %g cg %g c %g cond %g\n",
                rpnt->recnm1, rpnt->tspeed, rpnt->intenadj, rflux, pflux, 
		iflux, rpnt->flux, rpnt->pnois, rpnt->rhod, 
		rpnt->pde, rpnt->gcyc,
                 rpnt->cycg, rpnt->cond, rpnt->conduct);                /* */
 

         break;
        }       /* switch */
    }     /* for (rpnt; ; )  */
#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"...end\n");
#endif
}

/*------------------------------------*/

void runstim(double rtime, int dmsec, int endfl)

/* Advance all the stimuli to time "rtime",
 depending on their type and history.
 Also run all voltage and current clamps.
 Call at multiples of tenths of milliseconds only. */

{
#define RECTIM 1			/* how many 0.1 msec intervals */
#define READRECTIM 10			/* how often to read stimuli */
#define RTIMERND  0.00001		/* down-round value */
#define RTIMESTEP 0.0001		/* time step for reading stimuli */
#define QEFF 0.67			/* rhod isomerizations per photon */

   int i,nt;
   recstim *rspnt,*last,*next,*rbinpnt;
   recep *rpnt;
   node *npnt;
   comp *cpnt;
   static int readmsec=READRECTIM;
   static double start,stop,sens,rtimestep;

#ifdef DEBUG
  if (debug & 1)  fprintf (stderr,"runstim\n");
#endif

 rtimestep = max(timinc,RTIMESTEP);

/* The input from the file is sorted by time before being read in. 
   This is done as a final step by "stim".
*/

 if(readmsec++ >= READRECTIM) {
   if (!readstim(rtime+readmsec*rtimestep)) {  /* read new receptor inputs */
         fprintf (stderr,"runstim: invalid stimulus at time %g\n",rtime);
	 execerror ("Missing stimulus: "," stopping... ");
   }   
   readmsec = 0;                
 }

 if(dmsec >= RECTIM) {        /* calculate receptors for 0.1 msec intervals: */

	/* Problem here is that "rtime" comes from a continuing
	 * sum which has some roundup error, causing it to
	 * gradually increase to a greater value than the correct
	 * stimulus run time.  Therefore, we offset "rtime" and
	 * the "stop" time so a stimulus won't be missed. 
	 */
           
  /* start = rtime - (rtimestep + RTIMERND); */
  start = -1.0;					/* do and erase old stimuli */
  stop  = rtime + (rtimestep - RTIMERND);
  rtime -= RTIMERND;

 for (i=0; i<2; i++) {
  if   (i==0) rbinpnt = getstimp1(rtime);
  else        rbinpnt = getstimp2(rtime);
  if (endfl)  rbinpnt = recspnt;
  for (last=rspnt=rbinpnt; rspnt; ) {

/* fprintf(stderr,
 "stim num %d %d type '%c' val %g time %20.20g rtime %20.20g stop %20.20g\n",
 rspnt->recnm1,rspnt->recnm2,rspnt->ctype,rspnt->val,rspnt->time,rtime,stop);/**/
 
   if ((rspnt->time >= start) && (rspnt->time <stop)) {
     switch (rspnt->ctype) {

       case 'a':                /* absolute */
          if (findrecep (rspnt->recnm1, rspnt->recnm2, 
			 rspnt->recnm3, rspnt->recnm4,
					&rpnt, "runstim")) {

	   if (rpnt->ctype==TRANSDUCER) {
	      rpnt->flux = rspnt->val;
              if (cpnt=rpnt->comp1) {
		int nvclamps;

	       if (!(cpnt->miscfl & VBUF)) { /* not buf already */
                cpnt->extv = rpnt->flux;  /* set up voltage clamp */
                cpnt->v = rpnt->flux;    /* set comp voltage (for plot)*/
	        nvclamps = cpnt->miscfl & VEXT;
	        if (nvclamps==0) nvclamps++;	/* incr. num of vclamps */
	        if (nvclamps > VEXT) nvclamps=VEXT;
	        cpnt->miscfl &= ~VEXT;
	        cpnt->miscfl |= nvclamps;
	        }
	      }  /* if (npnt) */ 
	      else {
	        fprintf (stderr,"runstim: can't find compartment for node.\n");
	      }
	   }  /* if (rpnt->ctype==TRANSUCER) */

	   else {	/* rod or cone */
	    sens = rsens(rpnt,rspnt->wavel,rpnt->filt);
            rpnt->flux = rpnt->area * rspnt->val * sens * rpnt->attf * QEFF;
            if (rpnt->flux > MAXPHOT) rpnt->flux = MAXPHOT;
            if (rpnt->flux < 1e-10) rpnt->flux = 0.0;
	   }
	  }
#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim abs %d %g %g\n",
                                        rspnt->recnm1,rspnt->val,rspnt->wavel);
#endif
         break;

       case 'd':                /* delta */
          if (findrecep (rspnt->recnm1, rspnt->recnm2, 
			 rspnt->recnm3, rspnt->recnm4,
					&rpnt, "runstim")) {
	   if (rpnt->ctype==TRANSDUCER) {
	      rpnt->flux += rspnt->val;
              if (cpnt=rpnt->comp1) {
		int nvclamps;

	       if (!(cpnt->miscfl & VBUF)) { /* not buf already */
                cpnt->extv = rpnt->flux;  /* set up voltage clamp */
                cpnt->v = rpnt->flux;    /* set comp voltage (for plot)*/
	        nvclamps = cpnt->miscfl & VEXT;
	        if (nvclamps==0) nvclamps++;	/* incr. num of vclamps */
	        if (nvclamps > VEXT) nvclamps=VEXT;
	        cpnt->miscfl &= ~VEXT;
	        cpnt->miscfl |= nvclamps;
	        }
	      }  /* if (npnt) */ 
	      else {
	        fprintf (stderr,"runstim: can't find compartment for node.\n");
	      }
	   }  /* if (rpnt->ctype==TRANSUCER) */

	     else {	/* rod or cone */

	      sens = rsens(rpnt,rspnt->wavel,rpnt->filt);
              rpnt->flux += rpnt->area * rspnt->val * sens * rpnt->attf * QEFF;
              if (rpnt->flux > MAXPHOT) rpnt->flux = MAXPHOT;
              if (rpnt->flux < 1e-10) rpnt->flux = 0.0;
	     }
	  }
/* fprintf (stderr,"xwavel %g flux %g\n", rspnt->wavel,rpnt->flux);  */
/* fprintf (stderr,"absorb %g\n", rpnt->area * sens * rpnt->attf * QEFF); */

#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim delta %d %g %g\n",
                                        rspnt->recnm1,rspnt->val,rspnt->wavel);
#endif
         break;

       case 'e':                /* vclamp */
          if (npnt=findnode (rspnt->recnm1, rspnt->recnm2, 
			     rspnt->recnm3, rspnt->recnm4,
					 "runstim vclamp")) {
            if (npnt->comptr) {
		int nvclamps;

	     if (!(npnt->comptr->miscfl & VBUF)) { /* not buf already */
               npnt->comptr->extv = rspnt->val;  /* set up voltage clamp */
               npnt->comptr->v = rspnt->val;    /* set comp voltage (for plot)*/
	       nvclamps = npnt->comptr->miscfl & VEXT;
	       nvclamps++;			/* incr. num of vclamps */
	       if (nvclamps > VEXT) nvclamps=VEXT;
	       npnt->comptr->miscfl &= ~VEXT;
	       npnt->comptr->miscfl |= nvclamps;
	     }
	   }
	   else {
	      fprintf (stderr,"runstim: can't find compartment for node.\n");
	   }
          }
#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim vclamp %d %g\n",
                                        rspnt->recnm1,rspnt->val);
#endif
         break;

       case 'g':                /* cclamp */
          if (npnt=findnode (rspnt->recnm1, rspnt->recnm2, 
			     rspnt->recnm3, rspnt->recnm4,
				 	"runstim cclamp")) {
	    if (npnt->comptr) {
		int niclamps;
              npnt->comptr->exti += rspnt->val;    /* set up current clamp */
              npnt->comptr->miscfl |= IEXT;
	      niclamps = (npnt->comptr->miscfl & IEXT) >> 4;
	      niclamps++;			/* incr. num of iclamps */
	      if (niclamps > 15) niclamps=15;
	      npnt->comptr->miscfl &= ~IEXT;
	      npnt->comptr->miscfl |= (niclamps << 4);
	    }
	    else {
	      fprintf (stderr,"runstim: can't find compartment for node.\n");
	    }
          }

#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim cclamp %d %g\n",
                                        rspnt->recnm1,rspnt->val);
#endif
         break;

       case 'f':                /* volt clamp off */
          if (npnt=findnode (rspnt->recnm1, rspnt->recnm2,
			     rspnt->recnm3, rspnt->recnm4,
				"runstim voff")) {
	    if (npnt->comptr) {
		 int nvclamps;

	     if (!(npnt->comptr->miscfl & VBUF)) { /* not buf already */
	       nvclamps = npnt->comptr->miscfl & VEXT;
	       nvclamps--;			/* decr. num of vclamps */
	       if (nvclamps<=0) {
		  nvclamps = 0;
		  npnt->comptr->extv = 0.0;
	       }
	       npnt->comptr->miscfl &= ~VEXT;
	       npnt->comptr->miscfl |= nvclamps;
	     }
	    }
	    else {
	      fprintf (stderr,"runstim: can't find compartment for node.\n");
	    }
          }
#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim vclamp %d off\n", rspnt->recnm1);
#endif
         break;

       case 'h':                /* current clamp off */
          if (npnt=findnode (rspnt->recnm1, rspnt->recnm2,
			     rspnt->recnm3, rspnt->recnm4,
					"runstim voff")) {
	    if (npnt->comptr) {
		  int niclamps;

               npnt->comptr->exti -= rspnt->val;
	       niclamps = (npnt->comptr->miscfl & IEXT) >> 4;
	       niclamps--;			/* decr. num of iclamps */
	       if (niclamps<=0) {
			niclamps = 0;
/*             		npnt->comptr->exti = 0.0; */
		}
	        npnt->comptr->miscfl &= ~IEXT;
	        npnt->comptr->miscfl |= (niclamps << 4);
	    }
	    else {
	      fprintf (stderr,"runstim: can't find compartment for node.\n");
	    }
          }

#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"stim cclamp %d off\n",rspnt->recnm1);
#endif
         break;

       case 'i':                /* nt puff */
       case 'j':                
       case 'k':                
       case 'l':                
       case 'm':                
       case 'n':                
       case 'o':                
       case 'p':                
       case 'q':                
       case 'r':                
       case 's':                
	   nt = GLU + rspnt->ctype - 'i';
          if (npnt=findnode (rspnt->recnm1, rspnt->recnm2,
			     rspnt->recnm3, rspnt->recnm4,
				 	"runstim puff")) {
	    if (!setnt(npnt->comptr,nt,rspnt->val))
	      fprintf (stderr,"runstim: can't find compartment for node.\n");
          }
#ifdef DEBUG
          if (debug & 1)  fprintf (stderr,"puff %s %d %g\n",
                                        findsym(nt),rspnt->recnm1,rspnt->val);
#endif
         break;

        default:
         break;

      }         /* switch */

	if (endfl) next = rspnt->next;	/* next in stim list */ 	 
        else       next = rspnt->rnext;	/* next in bin */
        delrstim(rspnt);                /* delete stimulus after use */
        rspnt = next;
     }        /* if */

    else {   /* no stimulus here */
      last = rspnt;
      if (endfl) rspnt = rspnt->next;
      else       rspnt = rspnt->rnext;
    }
   }       /* for (rspnt=;;) */
  }       /* for (i;;) */
 }      /* if */
#ifdef DEBUG
  if (debug & 1)  fprintf (stderr,"runstim end\n");
#endif
}

 
