/* Module rndev in Program nc */

/* Generates binomial, Poisson, and gamma deviates (random sequences).
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifdef CPML
#include <cpml.h>
#else
#include <math.h>
#endif
#include <stdio.h>

#ifdef __cplusplus
}
#endif


#include "ncsub.h"

void restorstate(void);
char *emalloc(unsigned int n);
void *efree(void *ptr);

/*------------------------------------*/

double bnldev(double pp, register int n)
              
/* Taken from Press et al. 1988, "Numerical Recipes in C", p 223.
   Returns as a floating point number an integer value 
   that is a random deviate drawn from a binomial distribution
   of n trials each of probability pp, using drand() as a
   source of uniform random deviates.
*/

{
   register int j;
   register int bnl;
   static int nold = -1;  			     /* must be static */
   static double pold = -1, pc, plog,pclog,en,oldg;  /* must be static */
   static double am,em,g,angle,p,sq,t,y,z;
   double drand(), gammln2(double xx);
   void restorstate();

   if (pp > 1.0) pp = 1.0;
   if (pp < 0.0) pp = 0.0;
   p = (pp <= 0.5? pp : 1-pp);
   if (n < 50) {			/* use direct method if n is small */
      bnl = 0;
      for (j=1; j<=n; j++) 
	 if (drand() < p) bnl++;
   }
   else if ((am=n*p) < 1.0) {		/* if fewer than 1 event is expected */
      g = exp(-am);			/*  in 25 or more trials, then */
      t = 1.0;				/*  use Poisson method */
      for (j=0; j<n; j++) {
         t *= drand();
	 if (t < g) break;
      }
      bnl = j;
   }
   else {				/* use rejection method */
      if (n != nold) {
	 en = n;
	 oldg = gammln2(en+1.0);
	 nold = n;
      }
      if (p != pold) {			/* if p has changed */
	pc = 1.0 - p;
	plog = log(p);
	pclog = log(pc);
	pold = p;
      }
      sq = sqrt(2.0*am*pc);		/* rejection method with Lorentzian */
      do {				/* comparison function */
	do {
	   angle = drand();
	   if (angle == 0.5) angle = .49999999999;	
	   y = tan(MPI*angle);
	   em = sq*y+am;
	} while (em < 0.0 || em >= (en+1.0));	/* reject */
	em = floor(em);
	z = oldg-gammln2(em+1.0) - gammln2(en-em+1)+em*plog+(en-em)* pclog;
	if (z < -300.0) z = -300.0;	/* to fix range bug in exp() */
	t = 1.2 * sq * (1.0+y*y) * exp(z);

//	t = 1.2 * sq * (1.0+y*y) * exp(oldg-gammln2(em+1.0)
//		- gammln2(en-em+1) + em * plog+ (en-em) * pclog);

      } while (drand() > t);		/* reject. This happens about 1.5 */
      bnl = int(em);			/* times per deviate, on average. */
   }
   if (p != pp) bnl = n -bnl;
//   restorstate();			/* restore default random state  */
   return ((double)bnl);
}

/*------------------------------------*/

double fact(double f) {

  if (f > 1) return (f*fact(f-1));
  else       return 1;

}

/*------------------------------------*/

double gammln2(register double xx)

/* Approximation to the log(gamma) function */
/*  Table lookup is used. Table expands up to 10000 */

#define MAXGAMTABSIZ 20000

{
   register int x;
   static int runyet=0;
   static double y  = 0.5*log(2.0*MPI);
   double f,val;
   static int gamtabsiz = 0;
   static double *gamtab=(double*)NULL;
   double gammln(double xx);

 x = int (xx);
 if (x < MAXGAMTABSIZ) {		/* if too big, use approximation */
   if (x >= gamtabsiz) { 		/* make bigger table */
        register int i,j,mingam;
	double *ngamtab;
	static int ngamtabsiz;

    ngamtabsiz = int(xx/1024+1) * 2048;
    if (ngamtabsiz > MAXGAMTABSIZ) ngamtabsiz = MAXGAMTABSIZ;

  /* fprintf (stderr,"making new gamtab size %d\n",ngamtabsiz); */

    ngamtab = (double *)emalloc(ngamtabsiz*sizeof(double));
    if (gamtab) {			
      for (i=0; i<gamtabsiz; i++)	/* copy old table into new */
          ngamtab[i] = gamtab[i]; 
      for (; i<ngamtabsiz; i++)		/* fill rest with zeroes */
          ngamtab[i] = 0.0;
      efree(gamtab);
      mingam = gamtabsiz-1;
    } 
    else {
      ngamtab[0] = 0;
      ngamtab[1] = 0;
      mingam = 1;
    }
    gamtab = ngamtab;
    gamtabsiz = ngamtabsiz;
    for (i=mingam+1,f=j=i-1; i<gamtabsiz; f=j=i++) {
       gamtab[i] = gamtab[j] + log(f);
    }
  /* fprintf (stderr,"gamtab done\n"); */
  } 
  if (x < gamtabsiz) {			/* use table lookup */
    val = gamtab[x];
  }
 }
 else {
   val = (xx - 0.5) * log(xx) - xx + y + 0.083333754/xx;
 }				/* error corrected < 2e-10 for xx > 2000 */
 return val;
}

/*------------------------------------*/

double gammln(double xx)

/* Taken from Press et al. 1988, "Numerical Recipes in C", 2nd
   ed, p 214.  Returns the value ln(gamma(xx)) where xx > 0.
*/

{
   double x,y,tmp,ser;
   static double cof[6]={76.18009172947146,-86.50532032941677,24.01409824083091,
                        -1.2317395172450155, 0.1208650973866179e-2,
                        -0.5395239384953e-5};
   int j;

   y = x = xx;
   tmp = x + 5.5;
   tmp -= (x+0.5) * log(tmp);
   ser = 1.000000000190015;
   for (j=0; j<=5; j++) {
     ser += cof[j] / ++y;
   }
   return -tmp+log(2.5066282746310005*ser/x);
}

/*------------------------------------*/

int poisdev(double xm)
             
/*
 Return a poisson distribution which on average equals "xm".

 Taken from Press et al. 1988, "Numerical Recipes in C", p 222.
   Returns as a floating point number an integer value 
   that is a random deviate drawn from a Poisson distribution
   of mean xm as a source of uniform random deviates.
*/

{
  static double sq,alxm,g,oldm= -1.0;
  double em,t,y;
  double drand(),gammln(double xx);
  void restorstate();

#ifdef DEBUG
  if (debug & 1) fprintf (stderr,"poisdev\n");
#endif

  if (xm < 12.0) {			/* use direct method */
     if (xm != oldm) {
  	oldm = xm;
	g = exp(-xm);			/* if xm is new, compute exp */
     }
     em = -1;
     t = 1.0;
     do {
       em += 1.0;			/* instead of adding exp deviates */
       t *= drand();			/* we multiply uniform deviates */
     } while (t > g);
  }
  else {				/* use rejection method */
    if (xm != oldm) {
      oldm = xm;
      sq = sqrt(2.0*xm);
      alxm = log(xm);
      g = xm * alxm - gammln(xm+1.0);
    }
    do {
       do {
	  y = tan(MPI * drand());
	  em = sq*y+xm;
       } while (em < 0.0);		/* reject if zero probability */
       em = (int)(floor(em));
       t = 0.9 * (1.0+y*y) * exp(em*alxm-gammln(em+1.0)-g);
    } while (drand() > t);
  }
 // restorstate();			/* restore default random state */
 return (int)(em);
}

/*------------------------------------*/

double gamdev(int ia)

/* Taken from Press et al. 1992, "Numerical Recipes in C", p 292.
   Returns as a floating point number an integer value 
   that is a random deviate drawn from a gamma distribution
   of n trials each of order ia, using drand() as a
   source of uniform random deviates.
*/

{
   int j;
   double am,e,s,v1,v2,x,y;
   double drand(),gammln(double xx);

  if (ia < 1) ia = 1;
  if (ia < 6) {				/* use direct method, adding  */
     x = 1.0;				/* waiting times. */
     for (j=1; j<=ia; j++) x *= drand();
     x = -log(x);
  } else {			/* use rejection method */
   do {
      do {
         do {					/* These 4 lines generate */
	     v1 = 2.0*drand() - 1.0;		/* the tan of a */
	     v2 = 2.0*drand() - 1.0;		/* random angle */
         } while (v1*v1+v2*v2 > 1.0);
         y = v2/v1;
         am = ia - 1;
         s = sqrt(2.0 * am + 1.0);
         x = s * y + am;			/* Decide whether to reject x */
      } while (x <= 0.0);			/* Reject in reg.  of 0 prob */
      e = (1.0+y*y) * exp(am*log(x/am)-s*y);	/* Ratio of prob. fn. to comp */
   } while (drand() > e);			/* Reject on 2nd uniform dev. */
 }
// restorstate();			/* restore default random state */
 return x;
}

/*------------------------------------*/

double gasdev()

/* Taken from Press et al. 1992, "Numerical Recipes in C", p 289.
   Returns a normally distributed deviate with zero mean and
   unit standard deviation, using drand() as a source of 
   uniform random deviates.
*/

{
    static int iset=0;
    static double gset;
    double fac,rsq,v1,v2,retval;
    double drand();

  if (iset==0) {				/* Extra deviate ready ? */
    do {
        v1 = 2.0*drand() - 1.0;			/* Pick 2 uniform numbers */
        v2 = 2.0*drand() - 1.0;			/*  -1 to +1 in each dir */
        rsq = v1*v1 + v2*v2;			/* In unit circle ? */
    } while (rsq >= 1.0 || rsq == 0.0);         /*  not ? try again */

    fac = sqrt(-2.0*log(rsq)/rsq);

			/* now make Box-Muller transformation to get 2
			 * deviates.  Return one of them and save
			 * the other for next time */
/*    gset = v1 * fac;
    iset = 1;
*/
    retval = v2 * fac;
   }
/*
   else {
     iset = 0;
     retval =  gset; 
  }
*/
//  restorstate();			/* restore default random state */
  return retval;
}

/*------------------------------------*/

double rgasdev(int ngen)

/* Taken from Press et al. 1992, "Numerical Recipes in C", p 289.
   Returns a normally distributed deviate with zero mean and
   unit standard deviation, using drand() as a source of 
   uniform random deviates.
*/

{
    static int iset=0;
    static double gset;
    double fac,rsq,v1,v2,retval;
    double rrand(int ngen);

  if (iset==0) {				/* Extra deviate ready ? */
    do {
        v1 = 2.0*rrand(ngen) - 1.0;		/* Pick 2 uniform numbers */
	if (v1 < -10) return (0);		/* stop if wrong ngen */
        v2 = 2.0*rrand(ngen) - 1.0;		/*  -1 to +1 in each dir */
        rsq = v1*v1 + v2*v2;			/* In unit circle ? */
    } while (rsq >= 1.0 || rsq == 0.0);         /*  not ? try again */

    fac = sqrt(-2.0*log(rsq)/rsq);

			/* now make Box-Muller transformation to get 2
			 * deviates.  Return one of them and save
			 * the other for next time */
/*    gset = v1 * fac; */
/*    iset = 1; */
    retval = v2 * fac;
   }
/*   else {
     iset = 0;
     retval =  gset; 
  }
*/
  return retval;
}

/*------------------------------------*/
