/* symbol.c */

#include "nc.h"
#include "y.tab.h"

#define HASHSIZ 199                     /* prime number to make better hash */

static Symbol *hashtab[HASHSIZ] = {0};  /* symbol hash table */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <string.h>

#ifdef __cplusplus
}
#endif

char *emalloc(unsigned int n);
void efree(void *ptr);

/*---------------------------------------------------*/

void inithash(void)
{
   int i;

   for (i=0; i<HASHSIZ; i++) 
     hashtab[i] = 0;
}

/*---------------------------------------------------*/

int hashsym(char *s)
             

/* make disordered index of symbol name */

{
   int i;

   if (!s) return (0);
   for (i=0; *s; s++) i += *s;
   return (i%HASHSIZ); 
}

/*---------------------------------------------------*/

void hinstall (Symbol *sp)
              

/* Install a symbol in hash table */
/*  Uses "direct chaining" in a doubly-linked list. */
/*  Store latest entries at beginning of list */

{
    int i;
    Symbol *spnt,*slast;

   sp->next = 0;
   i = hashsym(sp->name);               /* initial index into hashtab */ 
   if (!(spnt=hashtab[i])) {              /* install directly in table */
      hashtab[i] = sp;
   }
   else {                               /* otherwise, go to end of list */
      sp->next = hashtab[i];
      hashtab[i] = sp;
   }
}

/*---------------------------------------------------*/

Symbol *lookup(char *s)      /* Find s in symbol table, given its name */
			     /*  Ignore STRING entries */                
{
        Symbol *sp;
        int i;
        
        i = hashsym(s);
        for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next) {
                if (sp->name) {
		        if (sp->type==STRING) continue;
			if (strcmp(sp->name, s) == 0) return sp;
		}
        }
        return (Symbol *)NULL;    /* 0 ==> not found */
}

/*---------------------------------------------------*/

Symbol *lookupnt(char *s, int t)/* find s in symbol table, given name and type */
			        /*  Ignore STRING entries */                
	      
{
        Symbol *sp;
        int i;
        
        i = hashsym(s);
        for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next) {
            if (sp->name) {
		if (sp->type==STRING) continue;
                if ((strcmp(sp->name, s) == 0) && (sp->type == t))
                        return sp;
	    }
        }
        return (Symbol *)NULL;    /* 0 ==> not found */
}

/*---------------------------------------------------*/

Symbol *lookupnta(char *s, int t, int starg) 
	/* find s in symbol table, given its name, type, and 
           beginning argnum for its { } block. */
	      
{
        Symbol *sp;
        int i;
        
        i = hashsym(s);
        for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next) {
                if ((strcmp(sp->name, s) == 0) && 
		    (sp->type == t) &&
		    (sp->argnum>=starg))
                        return sp;
        }
        return (Symbol *)NULL;    /* 0 ==> not found */
}

/*---------------------------------------------------*/

double lookpv(char *s)        /* Find symbol in symbol table, given its name */
			/*  and return its value */
                
{
        Symbol *sp;
        int i;

        i = hashsym(s);
        for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next) {
                if (sp->name)
                  if (strcmp(sp->name, s) == 0)
                        return sp->val;
        }
        fprintf (stderr,"lookupv: can't find '%s'\n",s);
        return 0;       /* not found ==> error, large num */
}

/*---------------------------------------------------*/

datum printsym()

{
	datum d = {0};
        Symbol *sp;
        int i,j;

    for (i=j=0; i<HASHSIZ; i++) {
        for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next,j++) {
                printf ("%s\n",sp->name);
	}
    }
    d.val = j;
    d.vtype = NUMBER;
    return d;
}

/*---------------------------------------------------*/

char *findsym(int num)              /* find a symbol's name, given its type */
                
{
        Symbol *sp;
        int i;

        for (i=0; i<HASHSIZ; i++) 
          for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next)
                if (sp->type == num)
                        return sp->name;
        return 0;       /* 0 ==> not found */
}

/*---------------------------------------------------*/

Symbol *findarg(int num)            /* find an argument given its number */
                

{
        Symbol *sp;
        int i;

        for (i=0; i<HASHSIZ; i++) 
          for (sp = hashtab[i]; sp != (Symbol *)0; sp = sp->next)
                if (sp->argnum == num)
                        return sp;
        return 0;       /* 0 ==> not found */
}

/*---------------------------------------------------*/

void erasarg(int argnum)           /* erase all arguments after func def */

{                               /*  erase all local variables, too. */
        Symbol *sp,*oldsp,*next;
        int i;

        for (i=0; i<HASHSIZ; i++) 
          for (oldsp = sp = hashtab[i]; sp != (Symbol *)0; sp = next)
            {
                next = sp->next;
/* if (sp->type==LOCALVAR) fprintf (stderr,"%s %d ",sp->name,sp->argnum); */
/* if (sp->type==ARG)fprintf(stderr,"%s %d %d * ",sp->name,sp->argnum,argnum);*/

                if ((sp->type == ARG || sp->type==LOCALVAR ||
			sp->type == LOCALARR) && sp->argnum >= argnum)
                 {
                   if (sp == hashtab[i]) hashtab[i] = next;
                   else oldsp->next = next;
                   if (sp->name) efree(sp->name);
                   efree(sp);
                 }
                else
                   oldsp = sp; 
           }
 /* fprintf (stderr,"\n"); */
}

/*---------------------------------------------------*/

Symbol *install(char *s, int t, double d)        /* install s in symbol table */
                
{
        Symbol *sp;

        sp = (Symbol *) emalloc(sizeof(Symbol));
        sp->name = emalloc(strlen(s)+1); /* +1 for '\0' */
        strcpy(sp->name, s);
        sp->type = t;
        sp->vtype = 0;
        sp->val = d;
        hinstall (sp);
/*      fprintf (stderr,"'%s' %ld\n",s,sp);     /* */
        return sp;
}

/*---------------------------------------------------*/

void erasymb(void)

/* erase symbol table */

{
   Symbol *next,*spnt;
   int i;
   static int called=0;

 if (!called) inithash();
 else {
   for (i=0; i<HASHSIZ; i++) 
     for (spnt=hashtab[i]; spnt; ) {
       next = spnt->next;
       efree (spnt->name);
       efree (spnt);
       spnt = next;
     } 
     hashtab[i] = 0;
 }
}


/*---------------------------------------------------*/

double getval(Symbol *sym)              
                    

/* return value of symbol given a pointer to symbol */

{ 
        return sym->val;
}

/*---------------------------------------------------*/

void setval(Symbol *sym, double val, int vtyp)         

/* set value of symbol, given its pointer. */

{ 
    if (sym) {
        if (vtyp==STRING)
	  sym->str = (char*)(long int)val;
	else
	  sym->val = val;
	if (vtyp) sym->vtype = (short int)vtyp;	/* if value type is set */
    }
}

        
