/* Module synfilt in program nc */

/* Synapatic filters */

#include <stdio.h>
#include "adef.h"
#include "nc.h"
#include "y.tab.h"
#include "ncsub.h"
#include "ncomp.h"
#include "control.h"

extern int cumlpfilt;

char *emalloc(unsigned int n);

void efree (void *ptr);

double ncabs(double val);

#ifdef __cplusplus
extern "C" {
#endif

#ifdef CPML
#include <cpml.h>
#else
#include <math.h>
#endif

#ifdef __cplusplus
}
#endif

/*------------------------------------*/

float *makfiltarr (int newsiz, int oldsiz, float *oarr, double val)

/* make larger filter array, copy old values to it */

{
    int i;
    float *narr;

  if (newsiz <= oldsiz) return((float *)NULL); /* do nothing if not larger */

  if ((narr=(float *)emalloc((newsiz)*sizeof(float))) == (float *)NULL) {
    fprintf (stderr,"makfiltarr: no space left for filter array\n");
    return ((float *)NULL);  
  }
  for (i=0; i<oldsiz; i++) {
    narr[i] = oarr[i]; 			/* copy the old values */
  }
  for (; i<newsiz; i++) {
    narr[i] = val; 			/* set the new spaces to val */
  }
  if (oarr) efree (oarr);		/* delete old filter array */
  return narr;
}

/*------------------------------------*/

float *copyfiltarr (int siz, float *oarr)

/* copy filter array to "save heap". */ 

{
    int i;
    float *narr;

  if ((narr=(float *)emalloc((siz)*sizeof(float))) == (float *)NULL) {
    fprintf (stderr,"copyfiltarr: no space left for filter array\n");
    return ((float *)NULL);  
  }
  for (i=0; i<siz; i++) {
    narr[i] = oarr[i]; 			/* copy the old values */
  }
  if (oarr) efree (oarr);		/* delete old filter array */
  return narr;
}

/*------------------------------------*/

double calctau(double tau, double timestep)

/* tau input in msec, output is multiplier for digital filter */
/* M.v.Rossum, 1996 */

{
   double k,taugran;
   int sign;

#define TIMERES  (.1)
#define MINTAU  (TIMERES)
#define TAUGRAN (.632120558)  /* (1 - 1/e) */

 sign = (tau >= 0);
 tau = ncabs(tau);
 if (tau < timestep*.1) tau =  timestep*.1;
 if (tau==0.0) tau = 1.0;

/*  taugran = timestep * pow(TAUGRAN,timestep/tau);  /* the old
 *  way */
/*  k = taugran / tau; */

 k = 1.0 - exp(-timestep/tau);		/* fix error for small tau*/
 if (k > 1.0) k = 1.0;			/* limit k to reasonable values */
 else if (k<=0.0) k = .0001;
 return ( sign ? k : -k);
}

/*------------------------------------*/

lpfilt *maksynfilt (int stype, int nfilt, double tfall, double val, 
		float *timec, double timestep)

/* make a time filter for a synapse */

{
     int i;
     lpfilt *fpnt;
     float *ftau,*lfilt;

  if (nfilt <= 0) return ((lpfilt*)NULL);
  if (nfilt > NUMFILT) nfilt = NUMFILT;
  nfilt++;				  /* make one more */
  if ((fpnt=(lpfilt *)emalloc(sizeof(lpfilt))) == (lpfilt *)NULL) {
      fprintf (stderr,"no space left for filter in synapse\n");
      return ((lpfilt *)NULL);  
  }
  lfilt = makfiltarr(nfilt, 0, (float*)NULL, 0.0);
  ftau  = makfiltarr(nfilt, 0, (float*)NULL, 0.0);
  fpnt->lfilt = lfilt;
  fpnt->ftau  = ftau;
  fpnt->ctype = LPFILT;
  fpnt->stype = stype;			 /* lp or hp */
  fpnt->nfilt = nfilt;			 /* number of filters */
  fpnt->tfall = tfall;			 /* lp fall tau or hp gain */
  for (i=0; i<nfilt-1; i++) {
 	 double timecn;
      if (timec) timecn = timec[i];	 /* check for missing timec */
      else       timecn = 1.0;
      fpnt->lfilt[i] = val;		 /* init low-pass filt array */
      fpnt->ftau[i] = calctau(timecn,timestep); /* time const, time step (ms)*/
  }
  fpnt->lfilt[i] = val;		 	/* init low-pass filt array */
  fpnt->ftau[i] = 0;			/* zero last one */
  cumlpfilt++;
  return (fpnt);
}

/*------------------------------------*/

double synfilt(lpfilt *lpnt, double transrate)

/* Make spnt->nfilt first-order low pass filters.
   Each filter has a time constant of spnt->ftau[i]/10 * stiminc (0.1 msec) */
{
   int i,prev,nfilt;
   double ctr,k,tr,oldtr;

  if (!lpnt) return transrate;
  switch (lpnt->ctype) {

  case LPFILT:
   tr=transrate;
   if (!(nfilt=lpnt->nfilt-1))		/* number of low pass stages   */
	return tr;			/*  leave one for output       */   
   if (lpnt->stype==HP) nfilt++;	/* if hp, always do at least 1 */
   oldtr = lpnt->lfilt[0];
   lpnt->lfilt[0] = tr;			/* zero filter holds input to filter */
   for (i=1; i<nfilt; i++) {		/* do 1 less filter than actual num */
     prev = i-1;
     k=lpnt->ftau[prev];
     if (synaptau > 0.0) k /= synaptau;	/* possibly modify time constant */
     if (k >= 1.0) k = .9999;		/* limit tau to reasonable value */
     ctr = lpnt->lfilt[i];
     if (k > 0) {
 	ctr += (tr - ctr) * k;   /* low pass filter */
     }
     else {				/* high pass */
        ctr += tr - oldtr;
	ctr += ctr*k;
     }
     oldtr = lpnt->lfilt[i];
     lpnt->lfilt[i] = tr = ctr; 
   }
		/* do the last one differently: separate falling phase */

   if (lpnt->stype==LP) {
     prev = i-1;
     k=lpnt->ftau[prev];
     if (synaptau > 0.0) k /= synaptau;	/* possibly modify time constant */
     if (k >= 1.0) k = .9999;		/* limit tau to reasonable value */
     ctr = lpnt->lfilt[i];
     if (lpnt->tfall>0.0) {		/* separate falling phase */
	 double trel;

       trel = (tr - ctr) * k; 	 	/* transmitter released */
       if (trel > 0) ctr += trel; 	/* rising phase, trel is non-neg */
       ctr *= lpnt->tfall;		/* falling phase */
       tr = ctr;
    }
    else {
      if (k > 0) {
 	ctr += (tr - ctr) * k;		/* low pass filter */
      }
      else {				/* high pass */
        ctr += tr - oldtr;
	ctr += tr*k;
      }
    }
   tr = ctr; 
   lpnt->lfilt[i] = tr;
   return (ctr);				/* LP: return last filter val */
  }	/* if LP */
  else return (ctr * lpnt->tfall);	/* HP: tfall is gain */
  break;

  case MFILT: return (0);
  break; 
  }

}

